/*
 * Goldwater IPC Handling
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.5 $  $Date: 2003/04/11 10:22:22 $
 *
 * -----------------------------------------------------------------------
 * $Log: bsipc.c,v $
 * Revision 1.5  2003/04/11 10:22:22  csmith
 * FreeBSD port still buggy!
 *
 * Revision 1.4  2003/04/10 20:44:40  csmith
 * changed __freebsd__ to FreeBSD__
 *
 * Revision 1.3  2003/04/10 20:35:16  csmith
 * ipcs handling for freebsd and linux
 *
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.10  2003/01/04 16:28:46  chris
 * gw_init is now intelligent about detecting if Goldwater is running.
 * By repeatedly calling gw_init() a client can make sure they stay attached
 * without having to do gw_init(), gw_release() over and over.  gw_init() will
 * detect that the backend has been bounced and try its hardest to reconnect.
 *
 * Revision 1.9  2003/01/03 00:09:07  chris
 * Use internal umask when creating files, shm and queues
 *
 * Revision 1.8  2003/01/02 20:25:30  chris
 * Allows configuration of IPC SHM umask (default is 066)
 *
 * Revision 1.7  2002/11/30 14:04:42  chris
 * GPL headers
 *
 * Revision 1.6  2002/06/16 11:39:51  chris
 * Changed BB Client list to be the GWDBridge Message Tracking list.
 *
 * Revision 1.5  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.4  2002/05/30 19:35:55  chris
 * Read maxclients from config and configure table in SHM
 *
 * Revision 1.3  2002/03/17 18:43:51  chris
 * Handle the destruction of semaphores and shm on exit
 *
 * Revision 1.2  2002/03/12 22:50:59  chris
 * Added BBReqQueue support.  Currently defines a BBReqQueue entry for
 * each server, even though servers are grouped into classes and thus
 * will usually share a queue.
 *
 * Revision 1.1  2002/02/17 22:31:29  chris
 * New sources - not all of these will be staying...
 *
 * -----------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdio.h>

#include "gw_ipcs.h"

#include <bsipc.h>
#include <gwlog.h>
#include <gwgbprm.h>


/* ----------------------------------------------------------------------- */

int
get_private_queue()
{
  int id = 0;

  id = msgget( IPC_PRIVATE, IPC_CREAT | (0666 & ~_Mask) );
  if( id < 0 ) {
    printf( "Cannot attach to (private) IPC\n" );
  }

  return id;
}


/* ----------------------------------------------------------------------- */

int
remove_queue( int qid )
{ 
  static struct msqid_ds bf;
  return msgctl( qid, IPC_RMID, &bf );
}

/* ----------------------------------------------------------------------- */
/*
 * Create a BBControl struct and Entry list in SHM, returning a pointer to
 * the BBControl struct.
 *
 * Returns NULL on SHM failure.
 */
BBControl *
create_bb( key_t ipckey, int max_serv, int max_svc, int max_accs, int max_dom )
{
  long  entry_size = 0;
  long  ctrl_size  = 0;
  long  svc_size   = 0;
  long  accs_size  = 0;
  long  dom_size   = 0;
  long  RQ_size    = 0;
  long  tot_size   = 0;
  int   ix         = 0;
  int   shmid      = 0;
  char *bbmem      = NULL;
  BBControl  *bbc  = NULL;
  BBDom      *bbd  = NULL;
  BBReqQueue *bbq  = NULL;

  ctrl_size  = shm_List_determine_memory_req( sizeof( BBControl  ), 1        );
  entry_size = shm_List_determine_memory_req( sizeof( BBEntry    ), max_serv );
  svc_size   = shm_List_determine_memory_req( sizeof( BBSvc      ), max_svc  );
  accs_size  = shm_List_determine_memory_req( sizeof( BBDMsgTrack), max_accs );
 
  dom_size   = sizeof( BBDom      ) * max_dom;  /* This is just an array */
  RQ_size    = sizeof( BBReqQueue ) * max_serv; /* This is just an array */

  tot_size = ctrl_size + entry_size + svc_size + accs_size + dom_size + RQ_size;

  if( (shmid = shmget( ipckey, tot_size, IPC_CREAT | IPC_EXCL | (0666 & ~_Mask)))  < 0 ) {

    if((shmid = shmget( ipckey, tot_size, (0666 & ~_Mask) )) < 0 ) {
      gw_logf( LOG_INTDBG, "shmget failure (%ld bytes)", tot_size );
      return NULL;
    }
  }

  /* shmat (really annoyingly!) returns -1 on failure, else a ptr
   * to the attached segment..... so checking for NULL failure return
   * fails. Argg!
   */
  if( (int)(bbmem = (char*)shmat( shmid, 0, 0 )) < 0 ) {
    gw_logf( LOG_INTDBG, "shmat failure" );
    return NULL;
  }

  /* Initialise Control Block
   */
  bbc = (BBControl*)bbmem;
  bbc->numbooted       = 0;
  bbc->servmax         = max_serv;
  bbc->svcmax          = max_svc;
  bbc->accessmax       = max_accs;
  bbc->dommax          = max_dom;
  bbc->queuemax        = max_serv;   /* MAX One queue entry per server */
  bbc->BBSize          = entry_size;
  bbc->BB_entry_offset = ctrl_size;
  bbc->BB_svc_offset   = ctrl_size + entry_size;
  bbc->BB_dtrack_offset= ctrl_size + entry_size + svc_size;
  bbc->BB_dom_offset   = ctrl_size + entry_size + svc_size + accs_size;
  bbc->BB_queue_offset = ctrl_size + entry_size + svc_size + accs_size + dom_size;

  bbc->spooldir[0]     = '\0';
  bbc->logname[0]      = '\0';
  bbc->loglevel        =  0;
  bbc->umask           =  066;

  bbc->gwhost[0]       = '0';
  bbc->gwport          =  0;

  bbc->maxmsgtx        = -1;
  bbc->domainid        = -1;

  uuid_clear( bbc->instance_id);

  /* 
   * Initialise BB lists
   */
  shm_List_init( (SHMLIST*)(bbmem + bbc->BB_entry_offset), 
		 sizeof( BBEntry), max_serv );
  
  shm_List_init( (SHMLIST*)(bbmem + bbc->BB_svc_offset), 
		 sizeof(BBSvc), max_svc );

  shm_List_init( (SHMLIST*)(bbmem + bbc->BB_dtrack_offset), 
		 sizeof(BBDMsgTrack), max_accs );

  /* Initialise domain array
   */
  bbd = (BBDom*)((char*)bbc + bbc->BB_dom_offset);

  for( ix = 0; ix < bbc->dommax; ix++ ) {

    bbd[ix].domainid  = ix;
    *(bbd[ix].domain) ='\0';
  }

  /* Initialise RequestQueue array
   */
  bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

  for( ix = 0; ix < bbc->queuemax; ix++ ) {

    bbq[ix].msgqid        = -1;
    bbq[ix].listen_ident = MAX_REQ_LISTEN_ID;
    bbq[ix].numonqueue   = 0;
    bbq[ix].aveQtime     = 0;
    bbq[ix].numdeps      = 0;
  }

  return bbc;
}

/* ----------------------------------------------------------------------- */

RetValue
destroy_bb( int shmid )
{
  RetValue ret = Success;

  if( shmctl( shmid, IPC_RMID, 0 ) < 0 ) ret = Failure;

  return ret;
}

/* ----------------------------------------------------------------------- */
/* IPC MSG Parameter Stuff
 */
#if defined(__FreeBSD__)
#  include "gw_ipcs_freebsd.c"
#endif

#if defined(__linux__)
#  include "gw_ipcs_linux.c"
#endif

/* ----------------------------------------------------------------------- */
/* end */
