#ifndef ADM_CMD_STATE
#define ADM_CMD_STATE
/*
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:10 $
 * 
 * --------------------------------------------------------------------------
 * $Log: command.h,v $
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.10  2003/03/26 19:28:14  chris
 * removed interactive from command list
 *
 * Revision 1.9  2002/11/30 20:00:01  chris
 * Work to allow forcing into interactive mode.
 *
 * Revision 1.8  2002/11/30 19:34:38  chris
 * Allow setting of administration prompt.
 *
 * Revision 1.7  2002/11/30 16:18:29  chris
 * Disable 'pager' for list commands in batch mode.
 *
 * Revision 1.6  2002/11/30 14:04:42  chris
 * GPL headers
 *
 * Revision 1.5  2002/10/20 13:38:30  chris
 * Enabled setting of loglevel from administrator (removed _LogLevel and
 * replaced it with *_LogLevelPtr through out code).
 *
 * Revision 1.4  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.3  2002/03/14 21:59:09  chris
 * Queue stats command
 *
 * Revision 1.2  2002/03/10 15:15:32  chris
 * Admin tool changes
 *
 * Revision 1.1  2002/02/26 00:30:20  chris
 * All new goldwater admin tool source.
 *
 * Revision 1.7  2001/12/26 22:57:42  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.6  2001/04/17 15:51:15  chris
 * Stats command and command list changes.
 *
 * Revision 1.5  2001/04/15 15:22:24  chris
 * Added svc_export, ldb_export and dom_join commands.
 *
 * Revision 1.4  2001/04/03 18:52:53  chris
 * TERMINATE command.
 *
 * Revision 1.3  2001/03/13 22:03:20  chris
 * Added 'Q' quit message
 *
 * Revision 1.2  2001/03/05 14:31:06  chris
 * Now handles the halting of servers (either by class, class/id or all).
 * Log command also works now.
 *
 * Revision 1.1  2001/01/14 18:38:56  chris
 * Initial revision
 *
 * --------------------------------------------------------------------------
 */

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>

#include <readline/readline.h>
#include <readline/history.h>

#define FALSE    0
#define TRUE     !FALSE
#define PLUSONE  (+1) 

#define TIMEOUT  "Timeout\n"
#define SUCCESS  "Success\n"
#define FAILURE  "Failure\n"
#define BUSY     "Busy\n"

#define PROMPT_MAX 20

typedef RetValue (CmndFunction)( CString * );
typedef enum { VAR_LONG, VAR_STRING } var_types;
typedef enum { V_LOGLEVEL, V_REQTOUT, V_PROMPT } VariableNames;

typedef struct {
  const char   *t_text;  /* Command Text            */
  CmndFunction *func;    /* Implementation Function */
  const char   *doc;     /* Help text for function  */
} CommandList;

typedef struct {
  const char    *t_text; /* Variable Text           */
  VariableNames  var;    /* Variable Name (enum)    */
  var_types      type;
  const char    *doc;    /* Help text for function  */
} VariableList;

void disable_pager( int p );

/* Callback Prototypes
 */
RetValue gw_help        ( CString * );
RetValue gw_boot        ( CString * );
RetValue gw_halt        ( CString * );
RetValue gw_cycle       ( CString * );
RetValue gw_process_list( CString * );
RetValue gw_service_list( CString * );
RetValue gw_domain_list ( CString * );
RetValue gw_queue_list  ( CString * );
RetValue gw_stats_list  ( CString * );
RetValue gw_quit_admin  ( CString * );
RetValue gw_set_command ( CString * );
RetValue gw_version     ( CString * );

#ifdef WANT_COMMAND_TABLE

/* Command lookup Table
 */
CommandList cmndtab[] = {
  { "help",    gw_help,         "Display this help page"                  },
  { "?",       gw_help,         "Synonym for help"                        },
  { "boot",    gw_boot,         "Launch Goldwater server instance(s)"     },
  { "halt",    gw_halt,         "Halt Goldwater server instance(s)"       },
  { "cycle",   gw_cycle,        "Cycle booted Goldwater servce instances" },
  { "srvl",    gw_process_list, "List booted servers"                     },
  { "svcl",    gw_service_list, "List service table"                      },
  { "doml",    gw_domain_list,  "List domain table"                       },
  { "queues",  gw_queue_list,   "List queue statistics"                   },

  { "stats",   gw_stats_list,   "Show current statistics"                 },
  { "quit",    gw_quit_admin,   "Exit administrator"                      },
  { "set",     gw_set_command,  "Allows parameter configuration"          },
  { "pl",      gw_process_list, "Synonym for srvl"                        },
  { "sl",      gw_service_list, "Synonym for svcl"                        },
  { "st",      gw_stats_list,   "Synonym for stats"                       },
  { "dl",      gw_domain_list,  "Synonym for doml"                        },
  { "ql",      gw_queue_list,   "Synonym for queues"                      },
  { "q",       gw_quit_admin,   "Synonym for quit"                        },
  { "version", gw_version,      "Give version information"                },
/*{ "interactive", NULL,        "Enter interactive mode from batch mode"  },*/
  { (char *)NULL, (CmndFunction *)NULL, (char *)NULL }
};


VariableList vartable[] = {
  { "loglevel", V_LOGLEVEL, VAR_LONG,  "Current Log Level"              },
  { "reqtout",  V_REQTOUT,  VAR_LONG,  "Request Timeout"                },
  { "prompt",   V_PROMPT,   VAR_STRING,"Administration Prompt"          },
  { (char *)NULL, 0, VAR_LONG, (char *)NULL }
};
#else

extern CommandList cmndtab[];
extern VariableList vartable[];

#endif

#endif

/* ------------------------------------------------------------------------ */
/* end */
