/*
 * Goldwater Distributed Web Architecture.
 * - Goldwater Boot Command
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.4 $  $Date: 2003/09/13 20:30:24 $
 * 
 * --------------------------------------------------------------------------
 * $Log: gw_boot_cmd.c,v $
 * Revision 1.4  2003/09/13 20:30:24  csmith
 * Better error message when service failes to execute.  Flush boot message so text up to -> appears whilst server is booting so user can see what is being waited for (this should always have been the case, but the flush was omitted for some reason :o)
 *
 * Revision 1.3  2003/09/13 13:04:32  csmith
 * Portable.net PInvoke helper functions.
 * Guardian bug: Did not clean BBReqQueue table on failure/shutdown of servers.
 * Increased version numbers to reflect significant bug fixes.
 * Announce copyright FSF in administrator.
 * Administrator now pauses 1/10 second between boot/halt commands to give
 * guardian time to breath (not strictly necessary, but may avoid some signal
 * race conditions).
 *
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.12  2003/03/25 21:16:20  chris
 * Enable auto-recycle of servers
 *
 * Revision 1.11  2002/11/30 14:04:42  chris
 * GPL headers
 *
 * Revision 1.10  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.9  2002/05/21 21:15:16  chris
 * Big tidy up - lots of misc changes.
 * Changed 'internal' services to be prefixed by '!' so they don't show in
 * the admin tool service list.
 *
 * Revision 1.8  2002/05/12 13:15:45  chris
 * Cleanup of unnecessary fields in BBEntry.
 *
 * Revision 1.7  2002/03/14 20:45:55  chris
 * Decent server halting!
 *
 * Revision 1.6  2002/03/11 22:34:30  chris
 * More uptodate TODO and other things.
 *
 * Revision 1.5  2002/03/10 15:15:32  chris
 * Admin tool changes
 *
 * Revision 1.4  2002/03/02 19:10:03  chris
 * Handle application and goldwater errors.
 *
 * Revision 1.3  2002/03/02 15:20:17  chris
 * Really does booting now.
 *
 * Revision 1.2  2002/02/28 00:56:13  chris
 * Wasn't initialising BB properly if detection failed for any reason.
 *
 * Revision 1.1  2002/02/26 00:30:20  chris
 * All new goldwater admin tool source.
 *
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_mmap.h>

#include <gwreq.h>
#include <gwlog.h>
#include <gwclient.h>
#include <gwapi.h>
#include <goldreq.h>
#include <regstrings.h>
#include <gwreadconf.h>

#include "gwadmin.h"

#define INV_CLASS  "ERR: Invalid Class Identifier\n"
#define INV_SERVER "ERR: Invalid Server Identifier\n"
#define BOOT_USAGE "Usage: boot -s server|-c class id -i service id|-c class id\
n\n"

static RetValue boot_server( long class, long instance );
RetValue launch_server( BootParams *bp );

extern char *local_dom;
extern char  local_dom_lev[];

/* ------------------------------------------------------------------------- */

RetValue
gw_boot( CString *argCS )
{
  long  val       = 0;
  CString optvCS = CSInit;
  char    opt    = '0';
  int     go     = 1;
  long    class  = 0;
  long    id     = 0;
  char    server[128] = {0};


  while( CSLenPtr( argCS ) > 0 ) {
    
    if( parse_arguments( argCS, &optvCS, &opt, "s:i:c:a" ) == Failure ) {
      /* Give Usage */
      printf( BOOT_USAGE );
      go = 0;
      break;
    }
    
    switch( opt ) {
    case 'c': /* Class */
      if( CStoInt32( &optvCS, &val ) == Failure ) {
	printf( INV_CLASS ); /* Error */
	go = 0;
      } else class = val;
      break;
      
    case 'i': /* ID */
      if( CStoInt32( &optvCS, &val ) == Failure ) {
	printf( INV_SERVER );
	go = 0;
      } else id = val;
      break;
      
    case 's': /* Server */
      if( CSLen( optvCS ) > sizeof(server) - 1 ) 
	CSLen( optvCS ) = sizeof(server) - 1;
      
      strncpy( server, CSStart(optvCS), CSLen(optvCS) );
      server[ CSLen(optvCS) ] = '\0';
      break;
      
    case 'a': /* Everything */
      break;
      
    default: 
      printf( BOOT_USAGE );
      go = 0;
      break;
    }
  }
  
  if( id >= 0 && class == -1 && server[0] == '\0' ) {
    printf( INV_CLASS );
    go = 0;
  }
  
  if( go ) boot_server( class, id );
  
  return Success;
}

/* ------------------------------------------------------------------------- */

static RetValue
boot_server( long class, long instance )
{
  long cix      = 0;
  long class_id = 0;
  long want_id  = instance;
  long len      = 0;
  int  co       = 0;
  int  lo       = 0;
  int  idx      = 0;
  int  sct      = 0;
  int  fct      = 0;
  
  long max      = 0;
  long min      = 0;
  long maxreq   = 0;
  char *server  = NULL;
  
  BootParams    bp;

  char path[128] = {0};

  printf( "\nInstalling server processes...\n\n" );

  /* Create config branch for classes... (local_dom_lev is a string %06ld)
   */
  co = sprintf( path, "%s%s%s", R_DOMAIN_B, local_dom_lev, R_SERVERS_SB );

  cix = 0;
  lo = sprintf( path + co, "%06ld", cix );
  strcpy( path + co + lo, R_ID_L );

  /* Loop for each class id in config
   */
  while( gw_config_read_long( path, &class_id ) >= 0 ) {
    
    strcpy( path + co + lo, R_SVC_MAX_L );
    if( gw_config_read_long( path, &max ) < 0 ) {
      return Failure;
    }

    strcpy( path + co + lo, R_SVC_MIN_L );
    if( gw_config_read_long( path, &min ) < 0 ) {
      return Failure;
    }

    strcpy( path + co + lo, R_EXE_L );
    if( gw_config_read_char( path, &server, &len ) < 0 ) {
      return Failure;
    }

    strcpy( path + co + lo, R_SVC_MAXREQ_L );
    if( gw_config_read_long( path, &maxreq ) < 0 ) maxreq = 0;

    bp.sp.servclass    = class_id;
    bp.sp.max          = max;
    bp.sp.min          = min;
    bp.sp.rstrt_thresh = maxreq;
    strcpy( bp.sp.server, server    );
    strcpy( bp.domain, local_dom );

    /* if all classes or the class we want 
     */
    if( !class || class == class_id ) {
      if( !want_id ) {
	/* We want all server instances of class 
	 */
	for( idx = 0; idx < min; idx++ ) {
          bp.sp.servid = class_id + idx;
          (launch_server( &bp ) == Success) ? sct++ : fct++;
	}
      } else {
	/* We want a single server instance - is it in the class we want? 
	 */
        if( want_id && (class_id <= want_id && want_id < class_id + max) ) {
          /* Boot specific instance */
          bp.sp.servid = class_id + (want_id - class_id);
          (launch_server( &bp ) == Success) ? sct++ : fct++;
        }
      }
    }
    
    if( class && class == class_id ) break; /* Done booting required class */
    
    cix++;
    lo = sprintf( path + co, "%06ld", cix );
    strcpy( path + co + lo, R_ID_L );
  } /* endwhile */

  printf( "\nStarted %d servers, %d failed\n", sct, fct );
  printf( "\n" );

  return Success;
}

/* ------------------------------------------------------------------------- */

RetValue
launch_server( BootParams *bp )
{
  GWMessage   msg     = {0};
  int         ret     = 0;
  RetValue    lret    = Failure;
  BootParams *rbp     = NULL;

  printf( "  ++ Booting %-15.15s Class: %-6ld ID: %-6ld -> ",
          bp->sp.server, bp->sp.servclass, bp->sp.servid );
  fflush(NULL);

  ret = gw_svc_call( "!/gctrl/srvboot", (char*)bp, sizeof(BootParams), &msg, 0 );

  rbp = (BootParams *)msg.data;

  if( ret >= 0 ) {
    printf( "Pid: %-ld ... %.*s\n", rbp->pid,(int)rbp->resp_len,rbp->resp_text);
    lret = Success;
  } else {

    if( gwerror == GWE_APPERR ) {
      printf( "Fail: %.*s\n", (int)rbp->resp_len, rbp->resp_text );
    } else {
      printf( "Fail: %s\n", gw_errorstr( gwerror ) );
    }
  }

  usleep( 100000 );

  gw_free_message( &msg );

  return lret;
}

/* ------------------------------------------------------------------------- */
/* end */
