/*
 * Goldwater IPC Handling
 *
 *
 * Copyright 1996-2002 netFluid Technology Limited
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF NETFLUID TECHNOLOGY LIMITED
 *
 * The copyright notice above does not evidence any actual or intended 
 * publication of such source code.    
 *
 *
 * -----------------------------------------------------------------------
 * $Log: gwipc.c,v $
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.16  2003/01/04 16:28:47  chris
 * gw_init is now intelligent about detecting if Goldwater is running.
 * By repeatedly calling gw_init() a client can make sure they stay attached
 * without having to do gw_init(), gw_release() over and over.  gw_init() will
 * detect that the backend has been bounced and try its hardest to reconnect.
 *
 * Revision 1.15  2003/01/03 00:09:09  chris
 * Use internal umask when creating files, shm and queues
 *
 * Revision 1.14  2002/08/01 18:15:26  chris
 * Domain ID fixes (incomplete)
 *
 * Revision 1.13  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.12  2002/04/14 00:33:50  chris
 * Allow booting of netserver (which is *slightly* different from normal
 * servers in that it does not need a message queue and does not have any
 * service).
 * remove_queue modified to only try and remove queues > 0
 *
 * Revision 1.11  2002/03/12 22:51:01  chris
 * Added BBReqQueue support.  Currently defines a BBReqQueue entry for
 * each server, even though servers are grouped into classes and thus
 * will usually share a queue.
 *
 * Revision 1.10  2002/02/24 18:08:59  chris
 * Appdir is no longer part of BootParams
 *
 * Revision 1.9  2002/01/03 20:27:31  chris
 * Introduction of preliminary Domain Control handling.
 *
 * Revision 1.8  2001/12/26 22:57:42  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.7  2001/06/03 10:48:48  chris
 * Removed attach_bb as it is also in gwlib.
 *
 * Revision 1.6  2001/04/15 12:55:03  chris
 * introduced domain table.
 *
 * Revision 1.5  2001/04/01 23:08:38  chris
 * Removed lrgflthrsh, init maxmsgtx to -1.
 *
 * Revision 1.4  2001/03/26 10:51:32  chris
 * Changed permissions to 0600.  attach_bb now fails if shmget fails,
 * previously wasn't testing shmget return value.
 *
 * Revision 1.3  2001/03/19 15:47:03  chris
 * Initialisation of new bbc parameters (logging, spooldir, maxmsgtx etc)
 *
 * Revision 1.2  2001/03/05 14:24:04  chris
 * Creates service map list.
 *
 * Revision 1.1  2001/01/14 18:22:49  chris
 * Initial revision
 *
 * -----------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <gwipc.h>
#include <gwgbprm.h>
#include <gwuuid.h>

/* ----------------------------------------------------------------------- */

int
get_queue( key_t key )
{
  int id = 0;

  id = msgget( key, IPC_CREAT | (0666 & ~_Mask) );
  if( id < 0 ) {
    printf( "Cannot attach to IPC\n" );
  }

  return id;
}


/* ----------------------------------------------------------------------- */

int
get_private_queue()
{
  int id = 0;

  id = msgget( IPC_PRIVATE, IPC_CREAT | (0666 & ~_Mask) );
  if( id < 0 ) {
    printf( "Cannot attach to (private) IPC\n" );
  }

  return id;
}


/* ----------------------------------------------------------------------- */

int
remove_queue( int qid )
{ 
  static struct msqid_ds bf;
  return (qid > 0) ? msgctl( qid, IPC_RMID, &bf ) : 0;
}

/* ----------------------------------------------------------------------- */
/* end */
