/*
 * Goldwater Distributed Web Architecture.
 *
 * Billboard Query Routines
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:10 $
 * 
 * --------------------------------------------------------------------------
 * $Log: bbquery.h,v $
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.15  2003/01/04 16:28:48  chris
 * gw_init is now intelligent about detecting if Goldwater is running.
 * By repeatedly calling gw_init() a client can make sure they stay attached
 * without having to do gw_init(), gw_release() over and over.  gw_init() will
 * detect that the backend has been bounced and try its hardest to reconnect.
 *
 * Revision 1.14  2002/09/01 10:59:31  chris
 * Addition of namespaceing to service name.
 * Service may now be of the form this.is.a.namespace/and/this/is/the/service
 * Wildcards are NOT allowed in RPC type message destinations.
 *
 * Revision 1.13  2002/08/04 19:36:03  chris
 * query_bb now allows scanning of a single domain's services.
 * API changes to allow replies to be sent back from dombridge.
 *
 * Revision 1.12  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.11  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.10  2002/03/17 18:44:07  chris
 * Handle the destruction of shm by exposing shmid value.
 *
 * Revision 1.9  2002/01/03 20:27:31  chris
 * Introduction of preliminary Domain Control handling.
 *
 * Revision 1.8  2001/12/26 22:57:42  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.7  2001/06/03 10:49:19  chris
 * Added attach_bb.
 *
 * Revision 1.6  2001/04/17 13:36:50  chris
 * All find functions take domainid.
 *
 * Revision 1.5  2001/04/15 12:54:45  chris
 * Introduced domain table handling.
 *
 * Revision 1.4  2001/03/26 18:46:37  chris
 * Parameters broken in srch_for_class_id
 *
 * --------------------------------------------------------------------------
 */
#ifndef _BBQUERY_H
#define _BBQUERY_H

#include <phlib/xl_types.h>
#include <goldreq.h>
#include <mutex.h>

RetValue   query_bb         ( const char *, BBSvc **, BBDom **,
			      SHMLIST *, long, long, long );
SHMLIST   *get_service_list ( key_t shmipc );
SHMLIST   *get_server_list  ( key_t shmipc );
void       release_bb_control();
BBControl *attach_bb        ( key_t shmipc, int *shmid );
BBControl *get_bb_control   ( key_t shmipc );

int        find_server_entry( long, long, long, BBEntry **, SHMLIST *);
int        srch_for_class_id( long, long, long, SHMLIST * );

int        find_services_for_server( long, long, long, const char *, 
				     BBSvc **, SHMLIST * );

/* NON MUTEXED FUNCTION CALL - ALWAYS WRAP MUTEX AROUND THIS CALL */
int        p_find_server_entry( long domainid, long class, long id,
		             BBEntry **s_ent, long *pos, SHMLIST *SRV_List );

BBDom *    __gw_find_domain( const char *, BBControl * );
BBDom *    __gw_find_domain_by_namespace( const char *, BBControl * );

#endif
