/*
 * Goldwater Distributed Web Architecture.
 *
 * User Command API
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.5 $  $Date: 2003/06/12 21:39:54 $
 * 
 * --------------------------------------------------------------------------
 * $Log: gwapi.c,v $
 * Revision 1.5  2003/06/12 21:39:54  csmith
 * Preliminary ASYNC request work.  Pnet C# wrappers.
 *
 * Revision 1.4  2003/04/12 18:09:54  csmith
 * SIGCONT didn't work under freeBSD, but SIGHUP did. Hmm.
 *
 * Revision 1.3  2003/04/10 18:11:22  csmith
 * FreeBSD port : change on_exit() to atexit()
 *
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.51  2003/03/25 21:31:13  chris
 * When checking cycle limit do reqcount % limit == 0 and not reqcount > limit !
 *
 * Revision 1.50  2003/03/25 21:16:24  chris
 * Enable auto-recycle of servers
 *
 * Revision 1.49  2003/03/24 22:12:32  chris
 * Allow forcing of disconnect (ignore reference count).
 *
 * Revision 1.48  2003/01/07 21:04:05  chris
 * Only do one sigcont as the second will kill the parent if running as root.
 *
 * Revision 1.47  2003/01/07 00:29:38  chris
 * Changed phlib linking as we've dropped /lib/phlib ext
 *
 * Revision 1.46  2003/01/05 22:26:34  chris
 * Lots of trickery to do superuser related tasks as quickly and safely as
 * possible before being permanently demoted to 'user'.
 *
 * Revision 1.45  2002/12/28 16:38:10  chris
 * When returning GWFailure from servers, return gwusrerr not gwerror.
 *
 * Revision 1.44  2002/11/11 20:26:34  chris
 * implemented gw_svc_forward.
 *
 * Revision 1.43  2002/08/27 18:52:31  chris
 * log quotes around server name
 *
 * Revision 1.42  2002/08/26 23:56:14  chris
 * Made sure log of request failure indicates that the server will exit....
 * This mislead me for a while!
 *
 * Revision 1.41  2002/08/26 18:57:56  chris
 * Store domain array idx in shm instead of the real msgqueue id.
 *
 * Revision 1.40  2002/08/24 13:23:16  chris
 * Broke __gw_complete_request into a sub call to __gw_complete_request_explicit
 * which allows the caller to explicitly define which service in which domain
 * the service counters are to be incremented for.
 *
 * Revision 1.39  2002/08/22 19:08:28  chris
 * Subtle logging changes
 *
 * Revision 1.38  2002/08/05 00:42:17  chris
 * Changed __gw_set_requestCtrl to take new and pass back copy of old - this
 * allows buffers to be swapped around the internal req_ctrl.
 * ie given parameters &A and &B, Perform: B <- req_ctrl <- A
 *
 * Revision 1.37  2002/08/04 19:36:03  chris
 * query_bb now allows scanning of a single domain's services.
 * API changes to allow replies to be sent back from dombridge.
 *
 * Revision 1.36  2002/08/01 18:15:26  chris
 * Domain ID fixes (incomplete)
 *
 * Revision 1.35  2002/07/28 18:27:12  chris
 * Further integration with netserver.
 *
 * Revision 1.34  2002/06/16 00:18:31  chris
 * Use gwuuid_compare to compare uuid_t buffers (4 32bit ints) because it is
 * faster.
 *
 * Revision 1.33  2002/06/15 19:32:05  chris
 * Message ID values are now 128 bit OSF DCE compatible uuid's.
 *
 * Revision 1.32  2002/06/11 21:52:58  chris
 * Enabled timed blocking in gwp_get_request and use this in dbridge to
 * prevent spinning and chewing cpu time.
 *
 * Revision 1.31  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.30  2002/06/04 16:35:45  chris
 * Introduce target_dom
 *
 * Revision 1.29  2002/06/02 19:26:42  chris
 * Maintain count of pending requests (requests in queue).
 *
 * Revision 1.28  2002/05/26 15:34:22  chris
 * Added GWC_FORWARD so the domain bridge is able to send the message
 * over the net without having to return a reply itself.
 *
 * Revision 1.27  2002/05/14 21:42:59  chris
 * Changes to allow gw_add_service to be done in the GWSRV_INIT
 * function of servers.
 * Most of this work was trying to find a way to set _MyServEntry in the
 * init stage of a function - however, in the case of guardian server, the
 * bbc->domainid parameter (specifically) is not set until the GWSRV_INIT
 * stage of its initialisation.  So setting _MyServEntry early does not
 * work for guardian, and thus cannot be done for other servers (rules for
 * one... and all that).
 *
 * Revision 1.26  2002/05/13 19:58:49  chris
 * Additional functionality to support the Domain Bridge.
 * Now allows __gw_get_request to 'not block' on request queue, and to
 * indicate that it returned because no messages were available.
 * More flags to control this behaviour (both control and private_user).
 *
 * Revision 1.25  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.24  2002/05/12 13:16:22  chris
 * Cleanup of unnecessary fields in BBEntry.
 * (Created new structure ServerParams that contains the fields common
 *  to those stored in the BB and those in the BootParam message).
 *
 * Revision 1.23  2002/05/12 11:17:43  chris
 * Block signals when in critical section of 'request' code (mainly
 * to stop mutex's being held in a claimed state - an issue with Guardian).
 * Problems are unlikely to occur with real servers.
 *
 * Revision 1.22  2002/05/11 13:40:44  chris
 * Clean up - for Domain handling.
 *
 * Revision 1.21  2002/04/21 20:19:51  chris
 * Fail if write block init failes.
 *
 * Revision 1.20  2002/04/14 00:02:26  chris
 * *** empty log message ***
 *
 * Revision 1.19  2002/04/12 20:10:02  chris
 * Domain ID's as stored in bbc->domainid are 1..n where as the actual
 * domain array is 0..n-1  When setting dptr, do so from bbc->domainid - 1
 *
 * Revision 1.18  2002/03/24 14:19:34  chris
 * It was trying to write 'Terminate' to _MyServerEntry when it was set to
 * NULL.
 *
 * Revision 1.17  2002/03/17 17:53:59  chris
 * When server is terminating, set the status in the BB to 'Terminating';
 *
 * Revision 1.16  2002/03/17 11:19:54  chris
 * No need to search for local domain in domain list as we know which
 * domain node is ours by looking at bbc->domainid;
 *
 * Revision 1.15  2002/03/12 22:51:03  chris
 * Added BBReqQueue support.  Currently defines a BBReqQueue entry for
 * each server, even though servers are grouped into classes and thus
 * will usually share a queue.
 *
 * Revision 1.14  2002/03/10 19:35:33  chris
 * Handles proper homed messages (initially for shutdown of servers).
 *
 * Revision 1.13  2002/03/02 19:09:39  chris
 * Handle skipping of stale messages in the queue.
 *
 * Revision 1.12  2002/02/24 18:48:48  chris
 * Tidied up callback handling.  Log startup of server.
 *
 * Revision 1.11  2002/02/24 18:05:38  chris
 * Moved APPDIR to BBControl block instead of duplicating it for each and
 * every server.
 * Return GWE_APPERR and set gwusrerr when Services return failure.
 * Use new regstrings that describe the new XML based configuration.
 * Added new functions to support individual addition of service entries.
 * New Version 1.14
 *
 * Revision 1.10  2002/01/29 22:43:40  chris
 * Cleaned up warnings.
 * Added gwp_ prefix to shutdown_server.
 * Added regdata.[ch] config entry to build.
 *
 * Revision 1.9  2001/12/26 22:57:43  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.8  2001/04/17 13:37:35  chris
 * Now local domainid aware.
 *
 * Revision 1.7  2001/04/01 22:40:50  chris
 * Nothing functional.  Comment corrections, removal of some set server
 * status calls.
 *
 * Revision 1.6  2001/03/26 18:43:50  chris
 * Implements the safe handling of 'homed' messages by re-transmitting
 * them if not intended for the current server and adjusting the 'listen'
 * mtype msgrcv value.
 * Implements on top of this architecture a controlled server shutdown
 * mechanism.  This is the only Homed message type at the moment.
 * TODO: Implement a context switcher for Homed messages.
 *
 * Revision 1.5  2001/03/26 10:47:50  chris
 * As we've transfered the memory cache info over to a GWMessage, all we
 * need to do it free this message when we come back from the Service
 * Function call.  If the user has re-used this message then the memory
 * will be free'd properly.  Thus req_ctrl.cache is given amniesia about
 * the memory, and its GWC_SPOOL flag is cleared so we don't get into a
 * mess with the free'ing responsibilities.
 *
 * Revision 1.4  2001/03/22 16:49:20  chris
 * __gw_get_request returns parameters back via its arguments (call by
 * ref).  These are needed by wrapper API's (eg Perl) as req_ctrl is not
 * exposed.
 *
 * Revision 1.3  2001/03/22 01:16:15  chris
 * Rework of the get_request function.  Split it into three parts.
 * Get the request, call the C/C++ service function, complete the
 * request.  This allows wrapper API's (such as Perl) to get the complete
 * the request, but provide their own function calling and loop code.
 *
 * Revision 1.2  2001/03/19 17:42:17  chris
 * CFG_* parameters (static const chars) were being included in multiple
 * files.
 *
 * Revision 1.1  2001/03/19 16:03:25  chris
 * Initial revision
 *
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <dlfcn.h>

#define GW_INTERNAL_FLAGS 1

#include <gwmsg.h>
#include <goldreq.h>

#include <gwportal.h>
#include <bbquery.h>

#include <gwapi.h>
#include <gwclient.h>
#include <gwgbprm.h>
#include <gwlog.h>
#include <gwtime.h>
#include <gwusers.h>

static const char SERV_STATE_IDLE [] = "Ready";
static const char SERV_STATE_CLOSE[] = "Terminate";

static const char SERVER_INIT_FUNC  [] = "GWSRV_INIT";
static const char SERVER_DEINIT_FUNC[] = "GWSRV_EXIT";

static GWMsgCtrl req_ctrl = {0};
static GWMessage reqmsg   = {0};
static int       _verbose = 0;
static int       _debug   = 0;

static long      _domainid = 0;
static int       _ReqQID  = 0;
static void     *sym_hand = NULL;

static long     ReqListenIdent = MAX_REQ_LISTEN_ID;

static void (*user_exit_func)() = NULL;

static void __gw_server_cleanup       ();
static int  __gw_set_server_state     ( BBEntry *, const char * );
static int  __gw_inc_server_done_count( BBEntry *s_ent );
static long __gw_re_tx_message        ( GWMsgCtrl *req_ctrl );
static long __gw_decrement_min_ident  ( );
static void __gw_initialise_user( const char* prog );

/* ------------------------------------------------------------------------- */
/*
 * Input:
 *
 * 1 - Domain
 * 2 - Server id
 * 3 - Class Id  (group almost)
 * 4 - Request Queue ID
 * 5 - IpcKey
 */
void
__gw_server_init( int ac, char **av )
{
  char   optstr[] = "vd:";
  char   c        = 0;
  char  *ep       = NULL;
  char **args     = NULL;
  GWStatus iret   = GWSuccess;

  GWStatus (*user_srv_init)  (char *, char * ) = NULL; /* Server name, unused */
  GWStatus (*user_srv_deinit)() = NULL;

  static int user_srv_init_check   = 0;
  static int user_srv_deinit_check = 0; 
  int parent_awake = 0;

  args = av;

  while( (c = getopt( ac, av, optstr )) > (-1) ) {

    switch( c ) {
    case 'v': /* Verbose */
      _verbose = 1;
      break;

    case 'd': /* Debug level */
      _debug = atoi( optarg );
      break;

    case ':':
      
      printf( "Missing Argument\n" );
      exit(1);
      break;

    case '?':
      printf( "Invalid Option\n" );
      break;
    }
  }

  if( ac - optind < 4 ) {
    printf( "Not enough control parameters...\n" );
    exit(2);
  }

  _Domain = av[optind++];

  _ServId = strtol( av[optind++], (char**)&ep, 10 );
  if( *ep != '\0' ) {
    fputs( "Invalid ServerID\n", stderr );
    exit(3);
  }

  _ServClass = strtol( av[optind++], (char**)&ep, 10 );
  if( *ep != '\0' ) {
    fputs( "Invalid ServerClass\n", stderr );
    exit(4);
  }

  _ReqQID = strtol( av[optind++], (char**)&ep, 10 );
  if( *ep != '\0' ) {
    fputs( "Invalid Queue ID\n", stderr );
    exit(5);
  }

  _IpcKey = strtol( av[optind++], (char**)&ep, 10 );
  if( *ep != '\0' ) {
    fputs( "Invalid IPCKey\n", stderr );
    exit(5);
  }

  _Mode = (char*)MODE_SHM; /* Servers are ALWAYS in SHM mode */
  
  /* Initialise Global Parameters (only needs _IpcKey )...
   */
  if( __gw_set_global_params() < 0 ) {
    fprintf( stderr, 
	     "Could not initialise startup params. Error %d\n", gwerror );
    exit(6);
  }

  /* Initialise log with serverid - We've got the logfile name from SHM now. */
  gw_log_init( _ServId, av[0] );

  /* Now that the log is open we can permanently demote oursleves to a normal
   * user (if we've a ruid of root) and do any chroot sandboxing.
   * -- This is a bit nasty.  If we're root and we demote to a real user
   *    permanently below, then we're not going to be allowed to send the
   *    parent (guardian) a signal saying that we have started.
   *    So we have to do it here.  This means that the user supplied init
   *    functions that return failures WILL NOT show up in the administrator.
   *    Or rather, they'll show up as having succeeded.  The price to pay for
   *    running as root, which is the price to pay for sandboxing (chroot).
   */
  if( !getuid() ) {
    /* We're running ruid superuser, so send the 'okay' signal whilst we've
	 * still got permission.
	 */
    usleep( 1000 );
    kill( getppid(), CHILD_OKAY_SIG );
	parent_awake = 1; /* Don't do the other CHILD_OKAY_SIG, parent will die if root */
  }

  __gw_initialise_user( av[0] );

  /* Get an entry into the symbol table for the current program...
   */
  if( !(sym_hand = dlopen( NULL, RTLD_LAZY | RTLD_GLOBAL )) ) {
    gw_logf( LOG_SERIOUS, "dlopen error '%s'", dlerror() );
    exit(6);
  }

  /* We cannot set up _MyServEntry here because bbc->domainid will NOT
   * be set until guardian finishes its GWSRV_INIT function.
   * That function is invoked next...
   * - Doing it here will work for all other servers except guardian... so
   *   we can't do it!
   */

  /* Now try and invoke the user supplied Server Init function (if there is
   * one...) - If this fails, then we need to log an error and NOT tell the
   * parent that we have started.
   * [ We get the init/deinit callbacks up front to avoid repeatedly hitting
   *   the symbol table ].
   */
  if( user_srv_init_check == 0 ) {
    dlerror(); /* Clear error status */
    user_srv_init = dlsym( sym_hand, SERVER_INIT_FUNC );
    if( dlerror() != NULL ) user_srv_init = NULL;
	user_srv_init_check = 1;
  }

  if( user_srv_deinit_check == 0 ) {
    dlerror(); /* Clear error status */
    user_srv_deinit = dlsym( sym_hand, SERVER_DEINIT_FUNC );
    if( dlerror() != NULL ) user_srv_deinit = NULL;
	user_srv_deinit_check = 1;
  }

  iret = GWSuccess;

  if( user_srv_init != NULL ) {
    /* Okay - launch user init function 
     */
    /* TODO: Create INFO struct containing server id, class etc and pass in..
     */
    iret = (*user_srv_init)( av[0], NULL );
  } 

  if( iret == GWSuccess ) {
    /* Tell guardian (or whoever the parent is...) that we're "okay to go"
	 * This will fail if we were running as ruid root, as we've been
	 * demoted.  We did a premeture kill above to get round this problem.
     */
    if(!parent_awake) {
      usleep( 1000 );
	  kill( getppid(), CHILD_OKAY_SIG );
    }

    gw_logf( LOG_INFO, "Server '%s' Started", av[0] );

		user_exit_func = (void*)user_srv_deinit;
		atexit( __gw_server_cleanup );

  } else {
    gw_logf( LOG_SERIOUS, "Server initialisation %s failed", SERVER_INIT_FUNC);
	exit(1); /* Get out - nothing more to do */
  }

  gw_log_init( _ServId, "" );

  return;
}


/* ------------------------------------------------------------------------- */

/*
static void
__gw_server_cleanup( int estatus, void *callback )
*/
static void
__gw_server_cleanup()
{

  /* Reset Status */
  if( _MyServEntry != NULL ) {
    __gw_set_server_state( _MyServEntry, SERV_STATE_CLOSE );
  }


  /* Call User Supplied De-Init function (if there is one)
   */
  if( user_exit_func != NULL) (*user_exit_func)(); 

  /* Remove any Reply Flag because we DO want the file unlinked.
   */
  req_ctrl.flags &= ~GWP_REPLY;
  gwp_cleanup_ctrl( &req_ctrl );

  gwp_remove_reply_queue();
}

/* ------------------------------------------------------------------------- */

static void 
__gw_initialise_user( const char* prog )
{
  uid_t u;
  gid_t g;

  /* We don't want to force down to the effective user if we're the
   * guardian process!
   */
  if( !prog || strcmp( prog, "guardian" ) == 0 ) return;

  /* If we're not really root, then nothing to do. 
   */
  if( getuid() ) return;

  if( __gw_become_real_user( &u, &g ) < 0 ) {
    fputs( "Could not become real user\n", stderr );
	exit(1);
  }

  /* ** DO ROOT SPECIFIC STUFF HERE LIKE CHROOT ** */

  /* Permanently discard the 'Root Hat'
   *
   * Set ruid, euid, saveuid to be euid 
   */
  if( __gw_become_user( u, g ) < 0 ) {
    fputs( "Could not setuid/setgid\n", stderr );
	exit(1);
  }
}

/* ------------------------------------------------------------------------- */
/*
 * Internal Service Function.
 * Sits and waits for a request data stream, which it passes to the requested
 * service function.
 *
 * Input:
 *  Ptr to flags:
 *    GWP_REQNBLK - Do not block waiting for request
 *                  Length returned as -1 if no request found in queue
 *                  (req_ctrl->cflags & GWC_NOREQM will be true in this case)
 *
 * Returns:
 *  Length of data
 *  Ptr to function name
 *  Ptr to Service name
 *  Ptr to request data
 *  Flags
 *
 *  - These parameters are available in the static req_ctrl struct in this 
 *    source file (used directly by the functions that comprise the request
 *    handling, keeping memory and spoolfiles cleaned etc).
 */
long
__gw_get_request( char **function, char **service, char **rdata, long *flgs )
{
  char  *data  = NULL;
  int    more  = 0;
  sigset_t ss, sso;

  gw_log_init( _ServId, "" );
  
  fflush(NULL);

  /* Block all signals - we don't want to be interrupted during a mutex loop 
   */
  sigfillset ( &ss );
  sigprocmask( SIG_BLOCK, &ss, &sso );

  do { 
    more = 0;
    
    if( _MyServEntry != NULL ) {
      /* Set server status */
      __gw_set_server_state( _MyServEntry, SERV_STATE_IDLE );
    }

    req_ctrl.cflags    = ((*flgs) & GWP_REQNBLK) ? GWC_NOBLOCK : 0; /*NOBLOCK*/
    req_ctrl.cflags   |= ((*flgs) & GWP_REQTBLK) ? GWC_TIMEBLK : 0; /*TIMEDBLK*/

    req_ctrl.msgqid    = _ReqQID;   /* Reset these values */
    req_ctrl.bsize     = _MaxMsgTx;
    req_ctrl.spool_dir = _SpoolDir;
    
	/* Sit and wait for requests - signals are not blocked because we want
	 * them to be detected asap.
	 */
    sigprocmask( SIG_SETMASK, &sso, NULL );

    if( (data = gwp_get_request( &req_ctrl, -(ReqListenIdent) )) == NULL &&
	   !(req_ctrl.cflags & GWC_NOREQM) ) {
      gw_logf( LOG_SERIOUS, "Get request FAILED - exiting" );
      gwp_finish_request( &req_ctrl ); /* Cleanup */
      exit(1);
    }

	/* block signals again.. */
    sigfillset ( &ss );
    sigprocmask( SIG_BLOCK, &ss, &sso );
  
    /* Get our server entry block in SHM
     *
     * - We can't do this until we've received a request message because the
     *   SHM probably won't be set up before then.....
     *   We only have to do this once though, so the extra hit just after
     *   receiving the first request is totaly acceptable.
     */
    if( _MyServEntry == NULL ) {
      /* Get reference to local domain stuff */
      BBDom *dptr = &((BBDom*)((char *)bbc + bbc->BB_dom_offset))[bbc->domainid - 1];
      
      _domainid = dptr->domainid;
      if( find_server_entry( _domainid, _ServClass, _ServId, 
			     &_MyServEntry, _ServerList ) < 0 ) {
        /* This is  s e r i o u s  and should never happen, but it has! */
	
        gw_logf( LOG_SERIOUS, 
		 "Internal Failure.  Cannot reference MyServEntry" );
	return -1;
      }
    }

    /* If no Request was received - leave returning -1 
	 */
    if( req_ctrl.cflags & GWC_NOREQM ) {
  
      sigprocmask( SIG_SETMASK, &sso, NULL ); /* Unblock of signals */
  
      return -1;
    }

    /* Now check the type of REQUEST message we've received.
	 * First consider HOMED messages
     */
    if( (req_ctrl.flags & GWP_HOMDMSG) && !(req_ctrl.cflags & GWC_NOREQM)) {
      /* This is a homed message */
      
      gw_logf( LOG_INTDBG, "DETECTED HOMED MESSAGE FOR %ld (I am %ld)", 
	           req_ctrl.destsrvid, _ServId );

      if( req_ctrl.destsrvid != _ServId ) {
	    /* But not for me.... re-send it
	     */
	    if( __gw_re_tx_message( &req_ctrl ) < 0 ) {
	      gw_logf( LOG_SERIOUS, "retx failure" );
	    }
	
      } else {
	    gw_logf( LOG_INTDBG, "Got homed message type: %x", 
		     req_ctrl.msgtype ); 
	    /*
	     * Consider different types of Homed Message....
	     */
	    switch( req_ctrl.msgtype ) {
	    case GWHM_HALT_SERVER: /* We're being asked to shutdown! */
	      gwp_shutdown_server();
	      gw_send_reply( GWSuccess, "", 0 );
	      exit(0); /* This is where we exit */
	      break;
	      
	    default:
	      break;
	    }
      }
      more = 1; /* Delt with message.  Loop round for next */
    }
  } while (more);
  
  gwerror = 0; /* Clear error code */

  /* We must have received a normal request message....
   */
  __gw_decrement_queue_count( _MyServEntry->req_queue_idx );

  /* Unblock of signals
   */
  sigprocmask( SIG_SETMASK, &sso, NULL );

  gw_log_init( _ServId, req_ctrl.service ); /* re-init with service name */
  
  if( gw_log_at_level( LOG_INTDBG ) ) {
    gw_logf( LOG_INTDBG, "__gw_get_request> Received Message" );
    
    if( req_ctrl.cflags & GWC_SPOOL ) 
      gw_logf( LOG_INTDBG, "__gw_get_request> Data in spoolfile" );
  }
  
  /* Set server status 
   */
  __gw_set_server_state( _MyServEntry, req_ctrl.service );

  /* Set return values
   */
  *service  = req_ctrl.service;
  *function = req_ctrl.function;
  *rdata    = CSStart( req_ctrl.cache );
  *flgs     = req_ctrl.flags;
  
  return CSLen( req_ctrl.cache );
}

/* ------------------------------------------------------------------------- */
/*
 * Returns a pointer to the current GWMsgCtrl data for the current request.
 * This is primarily for the Domain Bridge Server so that it is able to
 * cache it and receive further requests before actually dealing with them
 * (batch mode).
 */
GWMsgCtrl *
__gw_get_requestCtrl()
{
  return &req_ctrl;
}

/* ------------------------------------------------------------------------- */
/*
 * Sets the current GWMsgCtrl data for the current request.
 * It does this using a deep copy.
 *
 * This is primarily for the Domain Bridge Server so that it is able to
 * use gw_send_reply etc on GWMsgCtrl blocks that it has cached.
 *
 * old <- req_ctrl <- new 
 */
void 
__gw_set_requestCtrl( GWMsgCtrl *new, GWMsgCtrl *old )
{
  memcpy( (void*)old,       (void*)&req_ctrl, sizeof(GWMsgCtrl) );
  memcpy( (void*)&req_ctrl, (void*)new,       sizeof(GWMsgCtrl) );
}

/* ------------------------------------------------------------------------- */
/*
 * Invokes the specified C/C++ Goldwater Service (function)
 */
int
__gw_invoke_C_service()
{
  void (*svcfn)(char *, GWMessage * );

  svcfn = dlsym( sym_hand, req_ctrl.function );

  if( dlerror() != NULL ) {
    /*
     * SERVICE FUNCTION CANNOT BE FOUND!
     */
    gw_logf( LOG_SERIOUS, "Service Function '%s' cannot be resolved", 
	     req_ctrl.function );
    
    /*
     * TODO: Need to return error back to client...
     *       Do this by setting error flag(s) in req_ctrl
     *       initialising data component with error string
     *       and performing a service return.
     */
    return -1;
  }
  
  /*
   * Build a GWMessage struct to pass to the service
   */
  MirrorCString( reqmsg._bcx, req_ctrl.cache );
  
  reqmsg.data  = CSStart( req_ctrl.cache );
  reqmsg.len   = CSLen  ( req_ctrl.cache );

  if( req_ctrl.cflags & GWC_SPOOL ) {
    reqmsg._pflg |= GWC_SPOOL;
    strcpy( reqmsg._ctx, req_ctrl.spoolfile );
  } else reqmsg._pflg = 0;

  /* We've transfered our data ptrs over to reqmsg.  
   * This is all we need to be able to track it and clean it up once the
   * request has been completed.  If a user tries to use this reqmsg as a
   * return holder for a subsequent svc_call, then the memory will get re-used
   * sensibly.
   *
   * Clear ptrs and spoolfile flags in req_ctrl as they're now obsolete.
   */
  WrapCStringLen( req_ctrl.cache, "", 0 ); 
  req_ctrl.cflags &= ~GWC_SPOOL;
  
  gwerror = 0; /* Reset error code */
    
  if( gw_log_at_level( LOG_INTDBG ) ) {
    gw_logf( LOG_INTDBG, "Executing Service : '%s'", req_ctrl.service  );
    gw_logf( LOG_INTDBG, "          Function: '%s'", req_ctrl.function );
  }

  /* Call Service function
   */
  (*svcfn)( req_ctrl.service, &reqmsg );
  
  /*
   * Clean up the reqmsg (free memory, unmap file and unlink etc)
   */
  gw_free_message( &reqmsg );

  return 0;
}

/* ------------------------------------------------------------------------- */
/*
 * Finish the local request.
 * If done is true, will also update server and service counters.
 * This is performed against the current servers service set, within the
 * local domain and for the service recorded in the req_ctrl block.
 *
 * If you need to complete a request and update counters of a foreign domain
 * or servers services, then call __gw_complete_request_explicit instead.
 */
int
__gw_complete_request( int done )
{
  return __gw_complete_request_explicit( &req_ctrl, done, req_ctrl.service, bbc->domainid );
}

/* ------------------------------------------------------------------------- */
/*
 * This function completes a request and, if done is true, will update
 * the BB's server and service counts.
 *
 * This is explicitly called by GWDBridge over __gw_complete_request beacause
 * it needs to affect the counters of services in the local BB but recorded
 * against a REMOTE domain id.  Calling __gw_complete_request will only affect
 * services against the local domain.
 */
int
__gw_complete_request_explicit( GWMsgCtrl *rctrl, int done, const char *service, int domid )
{
  sigset_t ss, sso;
  BBSvc *sinfo = NULL;
  int    ret   = 1;

  /* Block signals
   */
  sigfillset ( &ss );
  sigprocmask( SIG_BLOCK, &ss, &sso );

  /* Check if the service sent a reply if it was supposed to!
   */
  if( done &&
      !(rctrl->flags & GW_NOREPLY) && 
	  !(rctrl->cflags & GWC_REPSENT) &&
	  !(rctrl->cflags & GWC_FORWARD) ) {
    gw_logf( LOG_SYSWARN, "Server did not dispatch an expected reply" );
  }
    
  /* Finish request and disconnect
   */
  gwp_finish_request( rctrl );
  gwp_portal_disconnect(); 
  
  if( done ) {
    if( find_services_for_server( domid, _ServClass, _ServId, 
				                  service, &sinfo, _ServiceList ) < 0 ) {
      gw_logf( LOG_SERIOUS,
	       "Cannot locate service '%s' in server: %ld:%ld,%ld for update",
	       service, domid, _ServClass, _ServId );
    } else {
      /* Update service done counter...
       */
      sinfo->reqdone++;
    }
  
    /* Increment count of number of requests processed and set server status
     * to 'reset'.
     */
    __gw_inc_server_done_count( _MyServEntry );
  } /* if done */

  sigprocmask( SIG_SETMASK, &sso, NULL ); /* Unblock Signals */

  /* Check to see if we're due to recycle */
  if( _MyServEntry->bb_sp.rstrt_thresh > 0 &&
	  (_MyServEntry->bb_sc.reqcount % _MyServEntry->bb_sp.rstrt_thresh) == 0) {

    /* Record that we're about to do a gracefull shutdown */
    _MyServEntry->bb_sc.gracerstrt_ind = 1;
	ret = 0;
  }
  
  return ret;
}

/* ------------------------------------------------------------------------- */
/*
 * Sends a reply to a waiting client...
 */
int
gw_send_reply( GWStatus status, const char *data, long len )
{
  GWMsgCtrl rctrl  = {0};
  RetValue  retv   = Success;
  long      bsent  = 0;
  long      left   = 0;
  long      tosend = 0;
  long      bsize  = 0;

  if( req_ctrl.cflags & GWC_REPSENT ) {
    /* Reply has already been sent...
     */
    gw_logf( LOG_SYSWARN, "Reply has already been sent to client" );
    gwerror = GWE_REPSENT;
    return -1;
  }

  if( req_ctrl.flags & GW_NOREPLY ) {
    /* No reply was expected...
     */
    gw_logf( LOG_SYSWARN, "Originator was not expecting a reply" );
    gwerror = GWE_REPNSNT;
    return -1;
  }

  if( data == CSStart(req_ctrl.cache) && len == CSLen(req_ctrl.cache) ) {
    gw_logf( LOG_INTDBG, "Reply appears to be the same buffer as the request");
  }
  
  /*
   * Make a copy of the request block, as there's info in it we need...
   * (queue id's etc)
   */
  memcpy( &rctrl, &req_ctrl, sizeof( GWMsgCtrl ) );

  /*
   * Completely replace the cache CString in our copy to stop the reply
   * trashing it the request cstring...
   */
  WrapCStringLen( rctrl.cache, "", 0 );
  
  rctrl.host = _GWHost;
  rctrl.port = _GWPort;

  rctrl.flags |= GWP_REPLY;
  rctrl.flags &= ~GWP_REMOTE;

  rctrl.total_size = len;

  /*
   * Check return status and mark reply.. (with special case for Domains)
   */
  if( status != GWSuccess ) {
    rctrl.flags |= gwusrerr;
    rctrl.flags |= GWE_SVCUFAIL;
    if( status == GWDomFailure ) rctrl.flags |= GWE_DOMFAIL;
  }

  if( gw_log_at_level(LOG_INTDBG) ) {
    char uuidbuf[64];
	uuid_unparse( rctrl.msgid, uuidbuf );
    gw_logf( LOG_INTDBG, "Sending reply with MSGID %s", uuidbuf );
  }

  /*
   * As we're sending a reply to a request, the request will have contained
   * the queue id to send the reply to.
   */
  rctrl.msgqid      = rctrl.r_msgqid; /* Reply Queue */
  rctrl.ident       = 1;

  /* Set the message type */
  rctrl.msgtype = (rctrl.flags & GWP_FROMDOM) ? GWM_DOM_REPLY : GWM_SVC_REPLY;

  bsize = rctrl.bsize;

  /*
   * Initialise flags and spoolfile stuff.....
   */
  if( gwp_write_block_init( &rctrl ) == Failure ) {
    gwerror = GWE_NOTCNCT;
    return -1;
  }

  /* If we're sending the reply to the Phoenix Goldwater portal (scheduler)
   * OR to a spoolfile, then do it all as one chunk.
   */
  if( (rctrl.cflags & GWC_PORTAL) || (rctrl.cflags & GWC_SPOOL) ) bsize = len;

  /*
   * Initiate timeout
   */
  signal( SIGALRM, __got_timeout );
  switch( sigsetjmp( timeout_env, 1 ) ) {
    
  case TIMER_START:
    alarm( _ReqTout );
    gw_logf( LOG_INTDBG, "gwapi.c: Sending reply to client" );

    bsent = 0;
    do {
      left   = len - bsent;
      tosend = left >= bsize ? bsize : left;
      
      if( gwp_write_block_data( &rctrl, data + bsent, tosend)== Failure) {
	gw_logf( LOG_SERIOUS, "Failed to send reply to client %ld [Error %ld]",
		 rctrl.clid, gwerror );
	
	rctrl.flags &= ~GWP_REPLY; /* Because we DO want the file unlinked! */
	gwp_cleanup_ctrl( &rctrl );

	signal( SIGALRM, SIG_IGN );
	alarm(0);
	return -1;
      }
      bsent += tosend;
    } while( bsent < len );
    
    retv = gwp_write_block_flush( &rctrl );

    signal( SIGALRM, SIG_IGN );
    alarm(0);

    if( retv == Failure ) {
      rctrl.flags &= ~GWP_REPLY; /* Because we DO want the file unlinked! */
      gwp_cleanup_ctrl( &rctrl );
      
      return -1;
    }
    break;
    
  case TIMER_EXPIRED:
    gw_logf( LOG_SYSERR, 
	     "Timeout occured when sending reply to client %ld", rctrl.clid );

    rctrl.flags &= ~GWP_REPLY; /* Because we DO want the file unlinked! */
    gwp_cleanup_ctrl( &rctrl );

    gwerror = GWE_TIMEOUT;

    return -1;
    break;
  }
  
  /* Cleanup ctrl block (which also means freeing cache memory and
   * unmapping/unlinking spool files
   */
  gwp_cleanup_ctrl( &rctrl );
  
  req_ctrl.cflags |= GWC_REPSENT; /* Indicate that reply has been sent */
  
  /* If we're returning a reply via GWDBridge, then we need to increment
   * GWDBridge's request queue as we're putting the reply there!
   */
  if( rctrl.flags & GWP_FROMDOM ) {
	/* Increment count of requests on this queue
	 */
    __gw_increment_queue_count( bbc->dbridge_msgq_idx );
  }
  /* TODO May have to increment ASYNC managers queue here if we're returning
   * a reply to the async manager
   */

  return 0;
}

/* ------------------------------------------------------------------------- */
/*
 * Forward data to another GW Service and then get out.
 *
 */
int
gw_svc_forward( const char *service, const char *reqdata, long reqlen, 
                long flags )
{
  GWMsgCtrl   ctrl  = {0};
  RetValue    retv  = Success;
  BBSvc      *sinfo = NULL;
  BBDom      *dinfo = NULL;
  BBReqQueue *bbq   = NULL;
  const char *svcname = NULL;
  char        namespace[ NAME_SPACE_MAX + 1 ] = {0};
  sigset_t    smask;

  long       bsent  = 0;
  long       left   = 0;
  long       tosend = 0;
  long       bsize  = 0;
  long       wantdom= 0;
  int        ix     = 0;

  if( _gw_initialised == 0 ) {
    gw_logf( LOG_SYSERR, "Not connected to Goldwater" );
    gwerror = GWE_NOTCNCT;
    return -1;
  }

  if( reqlen < 0 ) {
    gw_logf( LOG_SYSERR, "Bad request length" );
    gwerror = GWE_BADDATA;
    return -1;
  }

  gwerror = 0;

  /* Service may contain a namespace reference.  Extract and lookup if
   * necessary.
   *
   * my.namespace.declaration/some/service/name
   */
  wantdom = 0;
  ix = 0;
  /* Only scan service if we know it is a real service name and not a homed
   * data block (Which it will never be for a forwarded message).
   */
  while( !(flags & GWP_HOMDMSG ) && service[ix] > 0 && service[ix] != '/' && service[ix] != '!') ix++;
  svcname = service + ix;
  if( ix ) {
    if( ix > NAME_SPACE_MAX ) {
      gw_logf( LOG_SYSERR,"Namespace exceeds maximum length allowed (%d) [.*s]",
	       NAME_SPACE_MAX, NAME_SPACE_MAX, service  );
      gwerror = GWE_BADNSPC;
      return -1;
    }
    if( ix ) strncpy( namespace, service, ix );
    namespace[ix] = '\0';
    dinfo = __gw_find_domain_by_namespace( namespace, bbc );
    if( dinfo == NULL ) {
      gw_logf( LOG_SYSERR, "Unknown namespace [%s]", namespace );
      gwerror = GWE_BADNSPC;
      return -1;
    }
    wantdom = dinfo->domainid;
  }


  /* Keep only the flags that are valid for this api function.
   */
  ctrl.flags = req_ctrl.flags & ( GW_NOREPLY | GW_NOBLOCK | GWP_ASYNC );

  ctrl.org_r_msgqid= req_ctrl.org_r_msgqid; /* Reply message queue of origin */
  ctrl.total_size  = reqlen;
  ctrl.bsize       =  _MaxMsgTx;
  ctrl.ident       = req_ctrl.ident;

  /* Filled in later */
  ctrl.destsrvid   = -1; 
  ctrl.target_dom  = 1; 

  ctrl.clid        = req_ctrl.clid;
  uuid_copy( ctrl.msgid, req_ctrl.msgid );
  ctrl.host        = _GWHost;   /* Local Domain netServer Hostname/IP */
  ctrl.port        = _GWPort;   /* Local Domain netServer Port        */
  ctrl.spool_dir   = _SpoolDir;
  ctrl.service     = (char*)svcname;

  ctrl.msgtype = GWM_SVC_REQUEST;

  WrapCStringLen( ctrl.cache, "", 0 );

  if( query_bb( svcname, &sinfo, &dinfo,
		_ServiceList, _ServClass, _ServId, wantdom ) == Failure) {
    gw_logf( LOG_SYSWARN,
	     "Service not found or blocked (gwerror %ld)", gwerror );
    return -1;
  } 
  
  /* Get reference to BBReqQueue array */
  bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

  ctrl.msgqid   = bbq[ sinfo->req_queue_idx ].msgqid;
  
  /* Set MIN message Ident that all servers are looking for and get the
   * msgqid to send the request to...
   */
  
  /* TODO: Different values may be used here to provide some sort of 
   * priority to messages
   */
  ctrl.function   = sinfo->function;
  ctrl.target_dom = sinfo->domainid; /* Essential for dombridge routing */
  
  /* Increment count of requests on this queue
   */
  __gw_increment_queue_count( sinfo->req_queue_idx );
  
  if( sinfo->flags & GWS_REMOTE_SERVICE ) {
    /* Service has been found and is available on a remote goldwater
     * Domain.  The request for this queue will be the request queue
     * of the Domain Bridge.  We need to indicate the type of message
     * being sent to the Domain Bridge so that it knows the direction
     * of travel (request through bridge, or reply through bridge).
     */
    ctrl.msgtype = GWM_DOM_REQUEST;
  }

  if( gwp_write_block_init( &ctrl ) == Failure ) {
    /* gwerror will be set to a suitable error code */
    return -1;
  }

  /* Override REPLY message queue with that of the originator
   * _ReplyQueue will still be set to a valid reply queue for this server...
   *
   * TODO: Hide the setting of reply queue in a function in gwmsg.c 
   */
  ctrl.r_msgqid = req_ctrl.org_r_msgqid;

  /* TODO: block 'chunking' of data is no longer required as msgqueue messages
   *       are not 'chained'.
   *       This 'bsize' stuff and the loop can probably be removed.
   */

  /* If we're sending the reply to the NetServer TODO: Not Valid Any More!
   * OR we're spooling to disk, then do it all as one chunk.
   */
  bsize = ctrl.bsize;
  if( (ctrl.cflags & GWC_PORTAL) || (ctrl.cflags & GWC_SPOOL)) bsize = reqlen;

  /*
   * Initiate timeout
   */
  sigemptyset( &smask );
  sigaddset  ( &smask, SIGALRM );
  sigprocmask( SIG_UNBLOCK, &smask, NULL );

  signal( SIGALRM, __got_timeout );

  switch( sigsetjmp( timeout_env, 1 ) ) {
    
  case TIMER_START:
    alarm( _ReqTout );
    gw_logf( LOG_INTDBG, "gwapi.c: Setting service call timeout to %ld", 
	     _ReqTout );
    gw_logf( LOG_INTDBG, "gwapi.c: Sending request to service: %s",
	     ctrl.service);
                                           
    bsent = 0;
    do { 
      left   = reqlen - bsent;
      tosend = left >= bsize ? bsize : left;
      
      if( gwp_write_block_data( &ctrl, reqdata + bsent, tosend ) == Failure ) {
	gw_logf(LOG_SYSERR, "Failed to send request to Goldwater service %s",
		ctrl.service );

	ctrl.flags &= ~GWP_REPLY;
	gwp_cleanup_ctrl( &ctrl );

	signal( SIGALRM, SIG_IGN );
	alarm(0);
	return -1;
      }
      bsent += tosend;
    } while( bsent < reqlen );
    
    retv = gwp_write_block_flush( &ctrl );

    signal( SIGALRM, SIG_IGN );
    alarm(0);      

    if( retv == Failure ) {
      ctrl.flags &= ~GWP_REPLY; /* We want the spoolfile removed.. */
      gwp_cleanup_ctrl( &ctrl );
      return -1;
    }
    break;
    
  case TIMER_EXPIRED: 
    gw_logf( LOG_SYSERR, 
	     "Timeout occured when sending request to service: %s", ctrl.service );
    
    ctrl.flags &= ~GWP_REPLY;
    gwp_cleanup_ctrl( &ctrl );

    /* Decrement count of requests on this queue as it never got anywhere
     */
    __gw_decrement_queue_count( sinfo->req_queue_idx );
    
    gwerror = GWE_TIMEOUT;
    signal( SIGALRM, SIG_IGN );
    return -1;
    break;
  }

  /* Indicate that reply has been sent (in this case request passed on) */
  req_ctrl.cflags |= GWC_REPSENT;  
  
  /* If the original request came in across the domain then we need to 
   * increment GWDBridge's request queue count too.
   */
  if( req_ctrl.flags & GWP_FROMDOM ) {
    /* Increment count of requests on this queue
     */
    __gw_increment_queue_count( bbc->dbridge_msgq_idx );
  }

  return 0;
}

/* ------------------------------------------------------------------------- */

static int
__gw_set_server_state( BBEntry *s_ent, const char *str )
{
  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire BB Mutex" );
    return -1;    
  }

  strncpy( s_ent->bb_sc.state, str, SVC_NM_MAX );
  s_ent->bb_sc.state[SVC_NM_MAX] = '\0';

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release BB Mutex" );
    return -1;    
  }

  return 0;
}


/* ------------------------------------------------------------------------- */
/*
 * Increments server done counter.
 */
static int
__gw_inc_server_done_count( BBEntry *s_ent )
{
  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire BB Mutex" );
    return -1;    
  }

  s_ent->bb_sc.reqcount++;

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release BB Mutex" );
    return -1;    
  }

  return 0;
}

/* ------------------------------------------------------------------------- */

static long
__gw_re_tx_message( GWMsgCtrl *ctrl )
{
  RetValue  ret     = Success;
  long      txident = 0;

  /* Is is the first time this message has been picked up??
   *  If yes, then re-tx message with ident MIN(ident)
   *  Set LISTEN ident to be MIN(ident) - 1
   *  Set MIN(ident) to be MIN(ident - 1)
   * else
   *  ret-tx message with existing ident.
   *  Set LISTEN ident to be ident - 1;
   */
  if( ctrl->flags & GWP_TIAVM ) {

    ctrl->flags &= ~GWP_TIAVM;

    txident = __gw_decrement_min_ident(); /* Decrements by one and returns it*/
    if( txident <= 0 ) {
      gwerror = GWE_SYSFAIL;
      gw_logf( LOG_SERIOUS, "Could not calculate msg ident" );
      return -1;
    }
    ctrl->ident = txident + 1;
  } 

  ReqListenIdent = ctrl->ident - 1; /* same as stored in our listen_ident */

  gw_logf( LOG_INTDBG, "Re-sending message with MID %ld", ctrl->ident    );
  gw_logf( LOG_INTDBG, "Listening for requests on   %ld\n", ReqListenIdent );
  
  if( ret == Success ) ret = gwp_write_block_data ( ctrl, 
						    CSStart(ctrl->cache), 
						    CSLen  (ctrl->cache) );
  if( ret == Success ) ret = gwp_write_block_flush( ctrl );

  return (ret == Success) ? 0 : -1;
}

/* ------------------------------------------------------------------------- */

static long
__gw_decrement_min_ident()
{
  long   lident     = -1;
  BBReqQueue *bbq   = NULL;

  bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire BB Mutex" );
    return -1;    
  }

  lident = bbq[ _MyServEntry->req_queue_idx ].listen_ident;
  
  /*
   * Decrement min, and update our ident entry
   */
  if( lident > 0 ) {
    lident--;
    bbq[ _MyServEntry->req_queue_idx ].listen_ident = lident;
  }

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release BB Mutex" );
    return -1;    
  }

  return lident;
}

/* ------------------------------------------------------------------------- */
/* end */

