/*
 * Goldwater Distributed Web Architecture.
 *
 * config env value reader
 *
 * (c)2003 netFluid Technology Ltd - http://www.nfluid.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:11 $
 * 
 * --------------------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>

#include <gwreadconf.h>
#include <gwlog.h>
#include <gwdomain.h>
#include <regstrings.h>

static int find_local_domain_index();

/* ------------------------------------------------------------------------- */
/* read an ENV setting for the local domin in the goldwater config
 * As soon as the configuration is discarded, the pointers become invalid
 *
 */
const char *
gw_getenv( const char* env )
{
  int  ix        =  0;
  char path[128] = {0};
  long len       =  0;
  static int ldomlev = -1;

  char * val = NULL;

  if( ldomlev < 0 ) {
    ldomlev = find_local_domain_index();
	if( ldomlev < 0 ) return NULL;
  }

    /* Here we can scan the compiled Goldwater config in search of a
	 * 'env' setting in the local domain
	 */
    ix = 0;
	do {
	  /* Find env setting in GWConfig
	   */
	  sprintf( path, "/goldwater/domain/%06d/env/%06d/name", ldomlev, ix ) ;
      if( gw_config_read_char( path, &val, &len ) < 0 ) return NULL;

	  if( strcmp( val, env ) == 0 ) {
	    /* Found it, extract the associated value (string)
		 */
	    sprintf( path, "/goldwater/domain/%06d/env/%06d/env", ldomlev, ix ) ;
        if( gw_config_read_char( path, &val, &len ) < 0 || val[0] == '\0' ) {
		  return NULL;
	    }
		break;
	  }
	  val = NULL;
	  ix ++;
    } while(1);

    return val;
}

/* ------------------------------------------------------------------------- */
/* Returns local domain config index, or -1 if failure.
 */
static int
find_local_domain_index()
{
  char *local_dom = NULL;
  long  len       = 0;

  if( gw_config_read_char( R_LOCAL_DOM, &local_dom, &len ) < 0 ) {
      gw_logf( LOG_ERROR,  "Cannot determine local domain name from configuration\n" );
      return -1;
  }

  /* Find the domain branch node that the local domain may be found...
   */
  return __gw_find_domain_idx_in_config( local_dom );
}

/* ------------------------------------------------------------------------- */
/* end */
