/*
 * Goldwater Distributed Web Architecture.
 *
 * Multi-Chunk Data Encapsulation 
 *
 * Copyright 1996-2003 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.6 $  $Date: 2003/10/23 20:16:23 $
 * 
 * --------------------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <phlib/xl_malloc.h>
#include <phlib/xl_memory.h>

#include <gwmsgchain.h>

#define WORD_SIZE 4

static uchar_t * find_given_occ( MsgChain *msg, uint id, long occ );
static uchar_t * find_first_occ( MsgChain *msg, uint id );
static int       validate_buffer( MsgChain *msg );

/* ------------------------------------------------------------------------ */
/* Create a new message 
 *
 */
MsgChain *
gwmc_create()
{
  MsgChain *msg = NULL;
  uchar_t *p = NULL;

  /* Create Message 
   */
  if( (msg = (MsgChain*)xl_malloc( sizeof(MsgChain) ) ) == NULL ) return NULL;

  if( CreateCString( &msg->cstr, GWMC_MIN_MSG_SIZE ) == Failure ) {
    xl_free( (char*)msg );
    return NULL;
  }

  p = CSStart( msg->cstr );

  store_int32_LittleEndian( (p + GWMC_HDR_MAGIC_NUM_IX),   GWMC_MAGIC_NUMBER  );
  store_int32_LittleEndian( (p + GWMC_HDR_VERSION_IX  ),   GWMC_VERSION_V1_0  );
  store_int32_LittleEndian( (p + GWMC_HDR_BODY_OFFSET_IX), GWMC_HEADER_LENGTH );
  store_int32_LittleEndian( (p + GWMC_HDR_NUM_ITEMS_IX),   0                  );
  store_int32_LittleEndian( (p + GWMC_HDR_TOTAL_LEN_IX),   0                  );
   
  /* Move end pointer to start of body */
  CSLen( msg->cstr ) = GWMC_HEADER_LENGTH;

  msg->items = 0;

  /* Clear GWMessage holder so we don't try and clean it up later
   */
  memset( (char*)&(msg->gwmsg), 0, sizeof(GWMessage) );

  return msg;
}

/* ------------------------------------------------------------------------ */
/* Create a new message 
 *
 */
MsgChain *
gwmc_import( void * buffer )
{
  MsgChain *msg = NULL;
  uchar_t *p = NULL;

  /* Create Message Struct
   */
  if( (msg = (MsgChain*)xl_malloc( sizeof(MsgChain) ) ) == NULL ) return NULL;

  /* Just consider the header of the buffer for the moment (because we do
   * not know how many bytes are in the buffer)
   */
  WrapCStringLen( msg->cstr, buffer, GWMC_HEADER_LENGTH );

  /* Validate that this is a valid gwmessage
   */
  if( !validate_buffer(msg) ) return NULL;
   
  /* We can now extract the total length and initialise our struct 
   */
  p = (uchar_t*)buffer;

  CSLen(msg->cstr) = read_int32_LittleEndian(p + GWMC_HDR_TOTAL_LEN_IX);

  msg->items = read_int32_LittleEndian(p + GWMC_HDR_NUM_ITEMS_IX);

  /* Clear GWMessage holder so we don't try and clean it up later
   */
  memset( (char*)&(msg->gwmsg), 0, sizeof(GWMessage) );

  return msg;
}

/* ------------------------------------------------------------------------ */
/* Create a new message from a GWMessage
 *
 */
MsgChain *
gwmc_importGWMessage( GWMessage * gwmsg )
{
  MsgChain *msg = NULL;
  uchar_t *p = NULL;

  /* Create Message Struct
   */
  if( (msg = (MsgChain*)xl_malloc( sizeof(MsgChain) ) ) == NULL ) return NULL;

  /* Just consider the header of the buffer for the moment (because we do
   * not know how many bytes are in the buffer)
   */
  WrapCStringLen( msg->cstr, gwmsg->data, GWMC_HEADER_LENGTH );

  /* Validate that this is a valid gwmessage
   */
  if( !validate_buffer(msg) ) return NULL;
   
  /* We can now extract the total length and initialise our struct 
   */
  p = (uchar_t*)(gwmsg->data);

  CSLen(msg->cstr) = read_int32_LittleEndian(p + GWMC_HDR_TOTAL_LEN_IX);

  msg->items = read_int32_LittleEndian(p + GWMC_HDR_NUM_ITEMS_IX);

  /* Make a copy of the GWMessage so we can clean it up later
   */
  memcpy( (char*)&(msg->gwmsg), (char*)gwmsg, sizeof(GWMessage) );

  return msg;
}

/* ------------------------------------------------------------------------ */
/* Clear a message (not destroy it)
 */
void
gwmc_clear( MsgChain * msg )
{
  uchar_t *p = NULL;

  if( msg == NULL ) return ;

  gw_free_message( &(msg->gwmsg) ); /* Clear stored GWMessage if there is one */

  reset_cstring( &(msg->cstr) );  /* Clear but keep existing memory */
  p = CSStart( msg->cstr );

  /* Reset offsets etc */
  store_int32_LittleEndian( (p + GWMC_HDR_MAGIC_NUM_IX),   GWMC_MAGIC_NUMBER  );
  store_int32_LittleEndian( (p + GWMC_HDR_VERSION_IX  ),   GWMC_VERSION_V1_0  );
  store_int32_LittleEndian( (p + GWMC_HDR_BODY_OFFSET_IX), GWMC_HEADER_LENGTH );
  store_int32_LittleEndian( (p + GWMC_HDR_NUM_ITEMS_IX),   0                  );
  store_int32_LittleEndian( (p + GWMC_HDR_TOTAL_LEN_IX),   0                  );
   
  /* Move end pointer to start of body */
  CSLen( msg->cstr ) = GWMC_HEADER_LENGTH;

  msg->items = 0;

  /* Clear GWMessage holder so we don't try and clean it up later
   */
  memset( (char*)&(msg->gwmsg), 0, sizeof(GWMessage) );
}

/* ------------------------------------------------------------------------ */
/* Release message - free's all memory
 */
void
gwmc_destroy( MsgChain ** msg )
{
  __gwmc_destroy( *msg );
  *msg = NULL;
}

/* ------------------------------------------------------------------------ */

void
__gwmc_destroy( MsgChain *msg )
{
  if( msg == NULL ) return ;

  free_cstring( &(msg->cstr) );

  /* Clear stored GWMessage if there is one
   */
  gw_free_message( &(msg->gwmsg) );

  /* Discard memory */
  xl_free( msg );
}

/* ------------------------------------------------------------------------ */
/* Get a pointer to the data identified by ID and Occurance Number and
 * validate against content-type if supplied.
 *
 * If content-type is not supplied the content-type of the data item is
 * returned by the content-type ptr.
 *
 * returns a pointer to the data and the length.  Null if the ID/occurance
 * does not exist.
 */
void *
gwmc_read_data( MsgChain *msg, uint id, char ** content_type, int occ, long *len )
{
  int      want_occ = 0;
  long     dlen     = 0;
  uchar_t *occ_ptr  = NULL;
  uchar_t *data     = NULL;

  if( !validate_buffer(msg) ) return NULL;

  want_occ = occ ? occ : 1;

  if( (occ_ptr = find_given_occ( msg, id, occ )) == NULL ) {
    msg->error = GWMCE_OCC_NOT_FOUND;
    return NULL;
  }

  if( *content_type ) {
    /* Validate the content types match */
/* 	printf( "Check content type matches '%s'\n", *content_type ); */

  } else {
    *content_type = occ_ptr + GWMC_BLK_MIMETYPE_IX;
  }

  data = occ_ptr + read_int32_LittleEndian( occ_ptr + GWMC_BLK_DATA_OFFSET_IX );
  dlen = read_int32_LittleEndian( occ_ptr + GWMC_BLK_CONT_LEN_IX );

  *len = dlen;
  msg->error = GWMCE_NO_ERROR;

  return (void*)data;
}

/* ------------------------------------------------------------------------ */

int
gwmc_delete_data( MsgChain *msg, uint id, const char * content_type, int occ, long *len )
{
  if( !validate_buffer(msg) ) return -1;

  msg->error = GWMCE_NO_ERROR;
  return 0;
}

/* ------------------------------------------------------------------------ */
/* Store data of a given content-type in the message buffer along with an
 * identifier (numeric).
 * 
 * Returns the occurance number of the data if more than one item of the same
 * id exists.
 */
int
gwmc_add_data( MsgChain *msg, uint id, const char* content_type, const char* data, long len )
{
  int      blocklen = 0;
  int      totallen = 0;
  int      ctlen    = 0;
  int      olen     = 0;
  int      occ      = 1;
  long     last_block = 0;
  uchar_t *h        = NULL;
  uchar_t *p        = NULL;
  uchar_t *occ_ptr  = NULL;

  if( !validate_buffer(msg) ) return -1;

  ctlen = strlen( content_type ) + 1; /* +1 include null */

  /* Calculate block header len and round to next word boundary
   */
  blocklen = GWMC_BLK_SIZE + ctlen;
  blocklen = blocklen + (WORD_SIZE - blocklen % WORD_SIZE ) % WORD_SIZE;

  /* Calculate total message size and round to next word boundary
   * The +1 is there to leave enough space for a \0 terminator so we can
   * use strings directly in the buffer.
   */
  totallen = (blocklen + len + 1 ) + ( WORD_SIZE - (blocklen + len + 1) % WORD_SIZE ) % WORD_SIZE;

  /* Find the last occurance number of messages with this id....
   */
  if( (occ_ptr = find_given_occ( msg, id, -1 )) != NULL ) {
    occ = read_int32_LittleEndian( occ_ptr + GWMC_BLK_OCC_IX ) + 1;
  }

  /* Get the current block end (Which will be word aligned)
   */
  olen = CSLen( msg->cstr );
  last_block = olen;  /* This is the start of the new block - the new end */

  /* Make enough room in the buffer for the block header and the data itself
   * ( String length now encompasses header and data, though there is nothing
   *   in the new buffer space).
   */
  extend_cstring( &msg->cstr, totallen );

  /* Set the pointer to the current end of buffer
   */
  p = (uchar_t*)(CSStart( msg->cstr ) + olen);

  store_int32_LittleEndian( (p + GWMC_BLK_BLOCKLEN_IX), totallen );
  store_int32_LittleEndian( (p + GWMC_BLK_MSGID_IX   ), id);
  store_int32_LittleEndian( (p + GWMC_BLK_OCC_IX     ), occ);
  store_int32_LittleEndian( (p + GWMC_BLK_CONT_LEN_IX), len);

  /* There is no next occ yet */
  store_int32_LittleEndian( (p + GWMC_BLK_NEXT_OCC_IX), 0);  

  memcpy( (p + GWMC_BLK_MIMETYPE_IX), content_type, ctlen );

  store_int32_LittleEndian( (p + GWMC_BLK_DATA_OFFSET_IX), blocklen);

  /* Push current end of buffer to after block header */
  CSLen( msg->cstr ) += totallen;

  /* Append data */
  memcpy( (p + blocklen), data, len );
  (p + blocklen)[len] = '\0'; /* nul terminate data in buffer (not inc in len)*/

  /* Update last occurrance to point to this one
   */
  if( occ_ptr ) {
    store_int32_LittleEndian( (occ_ptr + GWMC_BLK_NEXT_OCC_IX), p - occ_ptr);
  }
 
  msg->items++;

  /* Keep message header up to date with the changes */
  h = CSStart( msg->cstr );
  store_int32_LittleEndian( (h + GWMC_HDR_NUM_ITEMS_IX), msg->items       );
  store_int32_LittleEndian( (h + GWMC_HDR_TOTAL_LEN_IX), CSLen(msg->cstr) );

  msg->error = GWMCE_NO_ERROR;
  return 0;
}

/* ------------------------------------------------------------------------ */
/* Return a pointer to the start of the message buffer and its length.
 * This is used so that the message may be sent over some transport or other.
 *
 * NOTE: The returned pointer si only valid while NOTHING else is done to 
 * the message.
 * Adding data or destroying message will leave the referenced data in an 
 * unpredictable state and subsequent use of it will probably cause a seg fault.
 */
void *
gwmc_get_message_ptr( MsgChain *msg, long *len )
{
  uchar_t *p = NULL;

  if( !validate_buffer(msg) ) return NULL;

  p = CSStart( msg->cstr );
  *len = read_int32_LittleEndian(p + GWMC_HDR_TOTAL_LEN_IX );
  return (void*)p;
}

/* ------------------------------------------------------------------------ */

void
gwmc_debug( MsgChain * msg )
{
  uchar_t *p = NULL;
  int blk = 0;

  if( !validate_buffer(msg) ) return;

  p = CSStart( msg->cstr );


  printf( "MessageChain at :%p\n", p );
  printf( "Header\n------\n" );

  printf( "Magic Number: %ld\n", read_int32_LittleEndian(p + GWMC_HDR_MAGIC_NUM_IX   ) );
  printf( "Version Num : %ld\n", read_int32_LittleEndian(p + GWMC_HDR_VERSION_IX    ) );
  printf( "Body Offset : %ld\n", read_int32_LittleEndian(p + GWMC_HDR_BODY_OFFSET_IX ) );
  printf( "Num Items   : %ld\n", read_int32_LittleEndian(p + GWMC_HDR_NUM_ITEMS_IX   ) );
  printf( "Total Length: %ld\n", read_int32_LittleEndian(p + GWMC_HDR_TOTAL_LEN_IX   ) );

  /* Step start pointer over header!!!!
   */
  p +=  read_int32_LittleEndian(p + GWMC_HDR_BODY_OFFSET_IX);

  printf( "\nNum Items   : %d\n", msg->items );

  while( p && blk < msg->items ) {
    printf( "Body Block %d\n-------------\n", blk );

    printf( "Block Length: %ld\n", read_int32_LittleEndian(p + GWMC_BLK_BLOCKLEN_IX   ) );
    printf( "Message ID  : %ld\n", read_int32_LittleEndian(p + GWMC_BLK_MSGID_IX      ) );
    printf( "Occurance   : %ld\n", read_int32_LittleEndian(p + GWMC_BLK_OCC_IX        ) );
    printf( "Next Occ Off: %ld\n", read_int32_LittleEndian(p + GWMC_BLK_NEXT_OCC_IX   ) );
    printf( "Content Len : %ld\n", read_int32_LittleEndian(p + GWMC_BLK_CONT_LEN_IX   ) );
    printf( "Data Offset : %ld\n", read_int32_LittleEndian(p + GWMC_BLK_DATA_OFFSET_IX) );
    printf( "MIME Type   : %s\n", (p + GWMC_BLK_MIMETYPE_IX   ) ); 

	p = p + read_int32_LittleEndian( p + GWMC_BLK_BLOCKLEN_IX );
    blk++;
  }
}

/* ------------------------------------------------------------------------ */

static uchar_t *
find_first_occ( MsgChain *msg, uint id )
{
  uchar_t *p  = NULL;
  uchar_t *op = NULL;
  int block_num = 0;

  if( !(msg && CSStart(msg->cstr) && CSLen(msg->cstr)) ) return NULL;

  p = CSStart( msg->cstr );
  p +=  read_int32_LittleEndian(p + GWMC_HDR_BODY_OFFSET_IX);

  while( p && 
		 block_num < msg->items && 
         read_int32_LittleEndian( p + GWMC_BLK_MSGID_IX ) != id ) {

#if 0
    printf( "\nFind id %d\nBody Block %d\n-------------\n", id, block_num );
#endif

	p = p + read_int32_LittleEndian( p + GWMC_BLK_BLOCKLEN_IX );
	block_num++;
  }

  if( p && block_num < msg->items && read_int32_LittleEndian( p + GWMC_BLK_MSGID_IX ) == id ) op = p;

  return op;
}

/* ------------------------------------------------------------------------ */
/*  occ < 0 will return the last occurrance.
 */
static uchar_t *
find_given_occ( MsgChain *msg, uint id, long occ )
{
  int occ_num = 1;
  uchar_t *p  = NULL;

  if( (p = find_first_occ( msg, id )) != NULL ) {
    while( read_int32_LittleEndian( p + GWMC_BLK_NEXT_OCC_IX ) && (occ < 0 || occ_num < occ) ) {
	  occ_num++;
	  p = p + read_int32_LittleEndian( p + GWMC_BLK_NEXT_OCC_IX );
    }
  }

  return (occ_num < occ) ? NULL : p;
}

/* ------------------------------------------------------------------------ */
/* Validate Buffer based on MAGIC NUMBER and VERSION
 */
static int
validate_buffer( MsgChain *msg )
{
  long     magic   = 0;
  long     version = 0;
  uchar_t *p       = NULL;
  
  if( !(msg && CSStart(msg->cstr) && CSLen(msg->cstr)) ) return 0;
  p = CSStart( msg->cstr );

  magic   = read_int32_LittleEndian(p + GWMC_HDR_MAGIC_NUM_IX );
  version = read_int32_LittleEndian(p + GWMC_HDR_VERSION_IX );

  msg->error = GWMCE_INVALID_MSG;

  if( magic != GWMC_MAGIC_NUMBER ) return 0;
  if( version > GWMC_VERSION_V1_0 ) return 0;

  msg->error = GWMCE_NO_ERROR;
  return 1;
}

/* ------------------------------------------------------------------------ */
/* end */
