/*
 * Config (Registry) Acess API
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.2 $
 *
 * --------------------------------------------------------------------------- 
 * $Log: gwreadconf.c,v $
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.10  2003/01/04 13:24:40  chris
 * Moved __gw_find_domain_idx_in_config from gwdomain to gwreadconf
 * as it is used in quite a few places outside libgwdom
 *
 * Revision 1.9  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.8  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.7  2002/04/14 00:03:44  chris
 * Added counter to init/deinit calls so that a call to the deinit func
 * does not release the config if someone else is referencing it.
 *
 * --------------------------------------------------------------------------- 
 *
 */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_memlst.h>
#include <phlib/xl_mmap.h>

#include <gwreadconf.h>
#include <gwlog.h>
#include <regstrings.h>

static CString bincfgCS = CSInit;

static int init_count = 0;

/* ------------------------------------------------------------------------- */
/*
 * Open Configuration file
 *  0  on success
 * -ve on failure:
 *     -1..-9 = Cannot open file
 *     -10    = Not a valid GW configuration file
 *     -11    = Configuration file not compiled for this platform
 */
int
gw_config_init()
{
  int      ret        = 0;
  char    *gwcfg      = NULL;
  CString  preambleCS = CSInit;

  gwerror = GWE_NOERROR;

  /* If config is open get out! */
  if( CSLen( bincfgCS ) ) { init_count++; return Success; }

  if( (gwcfg = getenv( "GWCONFIG" )) == NULL ) {
    gwerror = GWE_CFGRERR;
    return -1;
  }
  
  if( (ret = xl_mapfile( gwcfg, &bincfgCS )) < 0 ) {
    if( ret <= -2 ) gwerror = GWE_CFGNFND;
	else gwerror = GWE_BADCFG;
    return ret;
  }

  /* Make sure config file IS a Goldwater config file.... 
   */
  get_string_mid( &preambleCS, &bincfgCS, 0, 15   );
  string_right  ( &bincfgCS, CSLen(bincfgCS) - 16 );

  if( ! begins_with_chars( &preambleCS, "GWCFG" ) ) {
    gwerror = GWE_BADCFG;
    return -10;
  }

  if( *((int32*)(CSStart(preambleCS) + 12)) != 0xaabbccdd ) {
    gwerror = GWE_BCFGARC;
    return -11;
  }

  init_count++;

  return ret;
}

/* ------------------------------------------------------------------------- */

void
gw_config_deinit()
{
  init_count--;

  if( init_count == 0 && CSLen( bincfgCS ) ) { 
    xl_unmapfile( &bincfgCS );
    reset_cstring( &bincfgCS );
  }
  if( init_count < 0 ) init_count = 0; /* Just in case something slips */
}

/* ------------------------------------------------------------------------- */

static void *
find_flat_node( SHMLIST *stree, const CString *nodeCS, char *base )
{
  FlatTreeNode *fnode  = NULL;
  FlatTreeNode *ret    = NULL;

  if( shm_List_get_first( stree, (uchar**)&fnode ) == Success ) {
    do {
      if( fnode->namelen == CSLenPtr( nodeCS ) &&
	  strncmp( base + fnode->name_o, 
		   CSStartPtr(nodeCS), CSLenPtr(nodeCS) ) == 0 ) {
        ret = fnode;
	break;
      }
      
    } while( shm_List_get_next( stree, (uchar**)&fnode ) == Success );
  }
  
  return (void*)ret;
}


/* ------------------------------------------------------------------------- */

static SHMLIST *
traverse_flat_branch( CString *treeCS, CString *branchCS, FlatTreeNode **data )
{
  CString  tmpCS  = CSInit;
  CString  keepCS = CSInit;
  char    *base   = NULL;
  SHMLIST *stree  = NULL;
  void    *node;

  base  = CSStartPtr(treeCS);
  stree = (SHMLIST*)base;  /* Initial start tree */

  MirrorCString( keepCS, (*branchCS) );

  /* 
   * Special case - when the requested branch is '/ '
   */
  if( CSLen(keepCS) == 0 || 
      ( CSLen(keepCS) == 1 && CSChar(keepCS, 0) == '/' ) ) {
    if( shm_List_get_first( stree, (uchar**)data ) == Failure ) {
      *data = 0; /* Last node not found */
    } else {
      if( (*data)->ttype != T_BRANCH ) stree = 0;
    }
    return stree;
  }
  
  while( chop_line( &tmpCS, branchCS, '/' ) == Success ) {

    if( CSLen(tmpCS) == 0 ) continue;

    /* Move down tree */

    if( (node = find_flat_node( stree, &tmpCS, base )) != 0 ) { /* Found node */

      *data  = (FlatTreeNode*)node;

      if( ((FlatTreeNode*)node)->ttype != T_BRANCH ) return 0;
      if( (*data)->stree_o < 0 ) return 0; /* No Children! */

      /* Move _into_ node */
      stree = (SHMLIST*)(base + (*data)->stree_o);

    } else {
      /* node not found */
      *data = 0; /* Last node not found */
      MirrorCString( (*branchCS), keepCS );
      break;
    }
    MirrorCString( keepCS, (*branchCS) );
  }
  return stree;
}

/* ------------------------------------------------------------------------- */

int
gw_config_read_char( const char *path, char **data, long *len )
{
  int           ret      = -1;
  CString       branchCS = CSInit;
  CString       leafCS   = CSInit;
  FlatTreeNode *node     = 0;
  SHMLIST      *stree    = NULL;

  /* If we're not initialised, the do so!!!!!!
   */
  if(init_count == 0 && CSLen(bincfgCS) == 0 && gw_config_init() < 0) return -1;

  /* Keep a copy of the branch */
  WrapCString( branchCS, path );
  chop_line_last( &leafCS, &branchCS, '/' ); /* Chop off leaf */

  /* Make sure branch on which leaf is to be hung exists */
  stree = traverse_flat_branch( &bincfgCS, &branchCS, &node );
  
  /* Return failure if branch crosses a leaf node or branch incomplete */
  if( stree == 0 || node == 0 ) return -1; 
  
  node = (FlatTreeNode*)find_flat_node( stree, &leafCS, CSStart(bincfgCS) );
  
  if( node != 0 && node->ttype == T_LEAF )
  {
    *data = node->data_o + CSStart(bincfgCS);
    *len  = node->dlen;

    ret = 0;
  }

  return ret;
}

/* ------------------------------------------------------------------------- */

int
gw_config_read_long( const char *path, long *val )
{
  char *ptr = NULL;
  char *ep  = NULL;
  long  len = 0;

  if( gw_config_read_char( path, &ptr, &len ) < 0 ) return -1;

  *val = strtol( ptr, (char**)&ep, 10 );
  if( *ep != '\0' ) return -1;

  return 0;

}

/* ------------------------------------------------------------------------- */
/* This used to live in gwdomain.[ch] but is useful elsewhere, so we had to
 * put it into libgwtx
 */
int
__gw_find_domain_idx_in_config( const char *domain )
{
  int   idx       = 0;
  int   id        = 0;
  int   ret       = 0;
  long  len       = 0;
  char *data      = NULL;
  char  path[128] = {0};

  do {
    sprintf( path, "%s%06d%s", R_DOMAIN_B, idx, R_DOM_NAME_L );
    
    ret = gw_config_read_char( path, &data, &len );
    
    if( ret >= 0 && strcmp( domain, data ) == 0 ) {
      id = idx;
      break;
    }
  
    idx ++;
  } while( ret >= 0 );
 
  if( id < 0 ) {
    gw_logf( LOG_SERIOUS, "Domain %s missing from gwconfig", domain );
    return -1;
  }

  return id;
}

/* ------------------------------------------------------------------------- */
/* end */
