/*
 * Goldwater Distributed Web Architecture.
 *
 * Service Adv Handing
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.3 $  $Date: 2003/04/10 16:18:25 $
 * 
 * --------------------------------------------------------------------------
 * $Log: gwservice.c,v $
 * Revision 1.3  2003/04/10 16:18:25  csmith
 * Removed gwxml.h leftovers
 *
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.5  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.4  2002/05/26 18:15:15  chris
 * Added ServiceFlags to __gw_add_service
 *
 * Revision 1.3  2002/05/26 15:40:25  chris
 * Set flags against service to reflect whether it is local or not.  The
 * local boolean has been removed.
 *
 * Revision 1.2  2002/05/14 21:42:59  chris
 * Changes to allow gw_add_service to be done in the GWSRV_INIT
 * function of servers.
 * Most of this work was trying to find a way to set _MyServEntry in the
 * init stage of a function - however, in the case of guardian server, the
 * bbc->domainid parameter (specifically) is not set until the GWSRV_INIT
 * stage of its initialisation.  So setting _MyServEntry early does not
 * work for guardian, and thus cannot be done for other servers (rules for
 * one... and all that).
 *
 * Revision 1.1  2002/05/14 20:21:46  chris
 * Added gwservice.[ch] - service add/removal functions.
 *
 * --------------------------------------------------------------------------
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <goldreq.h>
#include <gwgbprm.h>
#include <mutex.h>
#include <gwlog.h>
#include <bbquery.h>

#include <gwservice.h>

#ifdef DEBUG
#  define DEBUGF gw_logf
#else
#  define DEBUGF if( 0 ) gw_logf
#endif

/* ------------------------------------------------------------------------- */
/*
 * Make sure you've locked the service table BEFORE calling this
 */
int
__gw_find_service_entry( long domainid, long classid, long srvid, 
			 SHMLIST *SVC_List )
{
  int    ix       = 0;
  uchar *ptr      = NULL;
  
  /* Find the collection of service ads for this server class......
   */
  ix = 0;
  if( shm_List_get_first( SVC_List, &ptr ) == Success ) {
    do {
      DEBUGF(  LOG_INTDBG, "Match domain ID %ld against %ld\n", domainid, ((BBSvc*)ptr)->domainid );
      if( ((BBSvc *)ptr)->domainid > domainid ) break; 
      if( ((BBSvc *)ptr)->domainid < domainid ) continue;

      /* Got domain... */
      DEBUGF( LOG_INTDBG, "Match servclass ID %ld against %ld\n", classid, ((BBSvc*)ptr)->servclass );
      if( ((BBSvc *)ptr)->servclass > classid ) break;
      if( ((BBSvc *)ptr)->servclass < classid ) continue;

      /* Got Domain and class... */
      DEBUGF( LOG_INTDBG, "Match servID %ld against %ld\n", srvid, ((BBSvc*)ptr)->servid );
      if( ((BBSvc*)ptr)->servid > srvid )  break;

      /* ix increment is in the while loop because we want to increase
       * ix whenever we do a continue (oh, and we have to do ++ix because
       * we need the increment to happen before we test its value as it'll be
       * ZERO (ie false) the first time around)..
       */
    } while( ++ix && shm_List_get_next( SVC_List, &ptr ) == Success ) ;
  }

  DEBUGF( LOG_INTDBG, "Found Service Table entry at pos %d\n", ix );

  return ix;
}

/* ------------------------------------------------------------------------- */

RetValue
__gw_add_service( long domainid, long classid, long srvid, long req_qidx,
                  SHMLIST *SVC_List, const char *service, const char *function,
		          long *qpos, ServiceFlags sflags )
{
  BBSvc   *svc_node  = NULL;

  const char *s = service;
  const char *f = function;

  if( ! *service ) s = f; /* If no service, use function name */

  if( strlen(f) > FNC_NM_MAX ) {
    gw_logf( LOG_SERIOUS,"Error adding service entry: Function name too long");
    return Failure;
  }
  if( strlen(s) > SVC_NM_MAX ) {
    gw_logf( LOG_SERIOUS,"Error adding service entry: Service name too long");
    return Failure;
  }

  DEBUGF( LOG_INTDBG, "Insert space in BB at position %ld for new service entry\n", *qpos);

  /* Create new service node with those for the current, class - or start
   * a new group at the start of the list
   *
   * Insert a new (empty) node
   */
  if( shm_List_insert_space( SVC_List, *qpos ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not add service '%s' to Service table", s );

    return Failure;
  }
      
  /* Get pointer to (empty) node
   */
  svc_node = (BBSvc *)shm_List_get_current_data( SVC_List );
    
  /* Fill in service details in node
   */
  strncpy( svc_node->service,  s, SVC_NM_MAX );
  svc_node->service[ SVC_NM_MAX ] = '\0';
    
  strncpy( svc_node->function, f, FNC_NM_MAX );
  svc_node->function[ FNC_NM_MAX ] = '\0';

  svc_node->domainid  = domainid;
  svc_node->servclass = classid;
  svc_node->servid    = srvid;
  svc_node->req_queue_idx = req_qidx;
  svc_node->reqdone   = 0;
  svc_node->picked    = 0;
  svc_node->flags     = (domainid != bbc->domainid ) ? 
                         GWS_REMOTE_SERVICE : GWS_LOCAL_SERVICE;
  svc_node->flags    |= sflags;

  (*qpos)++;

  DEBUGF( LOG_INTDBG, "Service added.  Next position is %ld\n", *qpos );

  return Success;
}


/* ------------------------------------------------------------------------- */
/*
 * Registers service with the calling servers Request Queue, but against
 * the supplied domain id
 */
int
__gw_add_service_dom( const char *service, const char *function, int domid )
{
  long pos        = 0;
  int  ret        = -1;
  BBEntry *bbe_p  = NULL; 

  static long qidx = -1; /* The servers request queue id */

  SHMLIST *BBS = (SHMLIST*)((char*)bbc + bbc->BB_svc_offset );
  SHMLIST *BB  = (SHMLIST*)((char*)bbc + bbc->BB_entry_offset );

  if( qidx < 0 ) {
    /* Find Server Entry in BB - and then get msqid from this
	 * We can't use MyServEntry because it may not be set up yet (if this
	 * function is called in GWSRV_INIT).
     */
    if( find_server_entry( bbc->domainid, _ServClass, _ServId, &bbe_p, BB) < 0){
      gw_logf( LOG_SERIOUS, "Could Not Find Server in BB (%ld:%ld:%ld)", 
        bbc->domainid, _ServClass, _ServId );
      return ret;
    }
    qidx = bbe_p->req_queue_idx;
  }   

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire SVC Mutex" );
    return ret;
  }

  /* Find Location of This Servers services within BB...
   */
  pos = __gw_find_service_entry( domid, _ServClass, _ServId, BBS );

  if ( __gw_add_service( domid, _ServClass, _ServId, qidx, BBS, 
	             service, function, &pos, 0 ) == Success ) ret = 0;

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release SVC Mutex" );
    ret = -1;
  }

  return ret;
}

/* ------------------------------------------------------------------------- */
/*
 * Public function to register services against the current server.
 */
int
gw_add_service( const char *service, const char *function )
{
  return __gw_add_service_dom( service, function, bbc->domainid );
}

/* ------------------------------------------------------------------------- */
/*
 * Removes services from service table for the server identified by 
 * class_id and server_id.
 * A server_id of -1 indicates that all services for the class are to be
 * removed.
 */
RetValue
__gw_remove_services( long domainid, long class, long id, SHMLIST *SVC_List )
{
  RetValue more          = Success;
  int      qp            = 0;
  int      stay_at_first = 0;
  uchar   *ptr           = NULL;

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire SVC Mutex" );
    return Failure;
  }
  
  /* Find the collection of service ads for this domain, class and optionally
   * server id......
   */
  qp = 0;
  if( shm_List_get_first( SVC_List, &ptr ) == Success ) {
    do {
      /* As we're deleting from a list that we're itterating through, we need
       * an external positional helper
       */
      stay_at_first = 0;

#if 0
      if( ((BBSvc *)ptr)->domainid > domainid ) break; /* not found */

      if( ((BBSvc *)ptr)->domainid  == domainid &&
	  ((BBSvc *)ptr)->servclass  > class )  break; /* not found */
      
      /* Hunt for class (sort by class second) */
      if( ((BBSvc *)ptr)->domainid  == domainid &&
	  ((BBSvc *)ptr)->servclass == class &&
	  ( id == -1 || ((BBSvc *)ptr)->servid == id ) ) 
#endif
      /* Try and find the same server id's (id third) */
      if( ((BBSvc *)ptr)->servclass == class &&
	  ( id == -1 || ((BBSvc *)ptr)->servid == id ) ) {
	
	/* Delete this list entry
	 */
	shm_List_delete_node( SVC_List );
	
	/* if at the first node, then we STAY examining the first node!*/
	if( qp == 0 ) stay_at_first = 1;
      } 
      
      if( stay_at_first ) more = shm_List_get_first( SVC_List, &ptr );
      else {
	qp++;
	more = shm_List_get_next( SVC_List, &ptr );
      }
    } while( more == Success );
  }

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release SVC Mutex" );
    return Failure;    
  }

  return Success;
}

/* ------------------------------------------------------------------------- */
/* end */

