/*
 * SHM Mutex (semaphores etc)
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:11 $
 * 
 * -------------------------------------------------------------------------
 * $Log: mutex.h,v $
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.10  2003/01/04 16:28:48  chris
 * gw_init is now intelligent about detecting if Goldwater is running.
 * By repeatedly calling gw_init() a client can make sure they stay attached
 * without having to do gw_init(), gw_release() over and over.  gw_init() will
 * detect that the backend has been bounced and try its hardest to reconnect.
 *
 * Revision 1.9  2002/05/30 19:37:34  chris
 * Create entries in BBClient table for each new client attaching to
 * application.
 *
 * Revision 1.8  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.7  2002/03/17 18:44:07  chris
 * Handle the destruction of shm by exposing shmid value.
 *
 * Revision 1.6  2002/02/28 00:55:16  chris
 * Various bug fixes.
 *
 * Revision 1.5  2001/12/26 22:57:43  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.4  2001/03/26 10:33:18  chris
 * ForceReset parameter on create_bbc_mutex
 *
 * Revision 1.3  2001/03/24 10:38:02  chris
 * Added get_bbc_mutex_value()
 *
 * Revision 1.2  2001/03/19 16:09:31  chris
 * Added proper logging.
 *
 * Revision 1.1  2001/03/13 22:30:39  chris
 * Initial revision
 * -------------------------------------------------------------------------
 */
#ifndef _SHM_MUTEX_H
#define _SHM_MUTEX_H

/*
 * Define Mutex identifier's and offsets
 */
#define MUTEX_BBC      0x0000
#define MUTEX_BBC_SVC  0x0001
#define MUTEX_BBC_CLT  0x0002
#define MUTEX_GUARDIAN 0x0003
#define MUTEX_SRV_BASE 0x0004

/* Number of system mutex's required
 */
#define NUM_SYS_MUTEX MUTEX_SRV_BASE


/* Function Prototypes
 */
RetValue create_bbc_mutex   ( key_t skey, int force_reset );
RetValue destroy_bbc_mutex  ( );
RetValue attach_bbc_mutex   ( key_t skey  );
void     detach_bbc_mutex   ( );
RetValue reset_all_bbc_mutex( );

RetValue aquire_bbc_mutex ( int sem_num );
RetValue release_bbc_mutex( int sem_num );

unsigned short get_bbc_mutex_value( int sem );

/* ------------------------------------------------------------------------- */
/* end */

#endif
