#ifndef PH_SESSION_H
#define PH_SESSION_H
/*
 * Session Handling
 *
 * Multiplexed Network Client/Server Core Engine
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * -----------------------------------------------------------------------
 * $Id: sessions.h,v 1.2 2003/03/29 12:21:11 csmith Exp $  $Date: 2003/03/29 12:21:11 $
 * -----------------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>

/* Number of fd's pending - Should be in select.h */
#ifndef howmany
#  define howmany(x,y) (((x)+((y)-1))/(y))
#  warning "defining HOWMANY"
#endif

/* Number of bits available for fd's (hence fd's available 
 * This parameter MUST be >= NOFILE (param.h)
 */
#ifndef FD_SETSIZE
#  define FD_SETSIZE 1024
#  warning "defining FD_SETSIZE 1024"
#endif

/* Number of file descriptor bits in a mask (long * 8bpb) */
#ifndef NFDBITS
#  define NFDBITS 32
#  warning "defining NFDBITS 32"
#endif

/*
 * Notes
 * -----
 * The maximum number of fd's available is given by FD_SETSIZE+NFDBITS
 * There is no way that you can have more fd's than this (once set).
 * Hence the total number of sockets and open file fd's must be less
 * than this value.
 *
 * The configuration below allows up to the MAX number of sockets available
 * and if all of these sockets are connected, then there will be no fd's
 * available for logging and writes etc.
 *
 * Be Warned 
 *
 * Events are not connected to fd's (other than to identify who ownes it)
 * and hence can be any size you want really.
 *
 */

#define MAX_FDS    FD_SETSIZE             /* Maximum fd's available */

#define RESERVE_FD 16   /* How many fd's to reserve for logging etc */

#define WR_PER_SES 10    /* Writes in queue per session */
#define EV_PER_SES 10    /* Events in queue per session */

#if 0 /* This is bad. We'll overrun our buffer if we hit fd > MAXFDS-RESERVE!*/
#define SES_MAX   (MAX_FDS - RESERVE_FD)     /* Max number of sessions       */
#endif

#define SES_MAX   (MAX_FDS-1)                /* Max number of sessions       */
#define WRB_MAX   ((SES_MAX+1) * WR_PER_SES) /* Max number of active wr fd's */
#define EVT_MAX   ((SES_MAX+1) * EV_PER_SES) /* Max number of active events  */

#define DEAD_SLOT 65536                      /* DEAD queue marker            */

/* Client Connection details
 */
typedef struct {
  int   ServiceIndex;
  void *PrivatePtr;
  void *SSLCon;
} NetworkConnectionData;

#endif
