/*
 * SSL support for Servers
 *
 * Multiplexed Network Client/Server Core Engine
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * -----------------------------------------------------------------------
 * $Id: xlssl.c,v 1.2 2003/03/29 12:21:11 csmith Exp $  $Date: 2003/03/29 12:21:11 $
 * -----------------------------------------------------------------------
 */

#include <stdio.h>

#include <sessions.h>

#ifdef USE_SSL

#include "xlssl.h"

/* ------------------------------------------------------------------------- */
/*
 * Handle SSL errors
 *
 * This may close the SSL session (the underlying TCP/IP transport layer may
 * still be open).
 */
int
handle_ssl_error( SSL **ssl, int err )
{
  switch( err ) {
    case SSL_ERROR_NONE:
       break;

	case SSL_ERROR_ZERO_RETURN:
	  break;

	case SSL_ERROR_WANT_WRITE:
	  break;

	case SSL_ERROR_WANT_READ:
	  break;

    case SSL_ERROR_WANT_CONNECT:
	  break;

	case SSL_ERROR_WANT_X509_LOOKUP:
	  break;

	case SSL_ERROR_SYSCALL:
	  break;

	case SSL_ERROR_SSL:
	  break;

  }

  xlssl_end_session( (unsigned char **)ssl );

  return err;
}

/* ------------------------------------------------------------------------- */

int
xlssl_begin_session( NetworkConnectionData *client_con, int fd, unsigned char *ssl_ctx )
{
  SSL_CTX *ctx = (SSL_CTX *)ssl_ctx;
  SSL     *ssl = NULL;
  int      err = 0;

  ssl = (SSL *)client_con->SSLCon;

  if( (ssl = SSL_new( ctx )) == NULL ) {
    /* Could not begin SSL connection */
  }

  SSL_set_fd( ssl, fd );

  while( (err = SSL_accept(ssl)) <= 0 ) {

    err = SSL_get_error( ssl, err );
	if( !(err == SSL_ERROR_WANT_READ || err == SSL_ERROR_WANT_WRITE) ) {
	  /* Serious Error 
	   */
      handle_ssl_error( &ssl, err ); /* This will CLOSE the SSL connection 
	                                  * but not the TCP/IP transport layer
			 		 	   		      */
      return -1 ;
    }
  }

  /* TODO: -- Check client certificater here -- */

  client_con->SSLCon = (uchar*)ssl; /* Return SSL connection */

  return 0;
}

/* ------------------------------------------------------------------------- */

void
xlssl_end_session( uchar **ssl_con )
{
  SSL *ssl = *(SSL **)ssl_con;

  SSL_free( ssl );
  *ssl_con = NULL;
}

/* ------------------------------------------------------------------------- */


#else /* NON SSL */


/* ------------------------------------------------------------------------- */

int
xlssl_begin_session( NetworkConnectionData *client_con, int fd, unsigned char *ssl_ctx )
{
  return -1; /* UNSUPPORTED */
}

/* ------------------------------------------------------------------------- */

void
xlssl_end_session( unsigned char **ssl_con )
{
  return;
}

/* ------------------------------------------------------------------------- */

#endif

/* end */
