/*
 * Goldwater Distributed Web Architecture.
 *
 * Tool for compiling Goldwater Configuration into a binary tree.
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:11 $
 * 
 * --------------------------------------------------------------------------
 * $Log: compileconf.c,v $
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.18  2003/03/25 19:51:03  chris
 * added maxrequests
 *
 * Revision 1.17  2003/01/05 23:00:56  chris
 * Allow sandbox option in config.
 *
 * Revision 1.16  2003/01/02 20:25:33  chris
 * Allows configuration of IPC SHM umask (default is 066)
 *
 * Revision 1.15  2003/01/02 18:49:56  chris
 * Added User and Group
 *
 * Revision 1.14  2002/12/21 21:48:43  chris
 * Added <env .. /> to <domain> element
 *
 * Revision 1.13  2002/09/01 11:24:38  chris
 * Changed version number to 1.4
 *
 * Revision 1.12  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.11  2002/05/30 19:37:52  chris
 * Added support for maxclients in config.
 *
 * Revision 1.10  2002/05/26 18:03:22  chris
 * export is now an attribute of the service tag.
 *
 * Revision 1.9  2002/05/26 17:54:08  chris
 * Added export tag
 *
 * Revision 1.8  2002/05/26 17:22:54  chris
 * Added import tag to domain section
 *
 * Revision 1.7  2002/05/21 21:15:16  chris
 * Big tidy up - lots of misc changes.
 * Changed 'internal' services to be prefixed by '!' so they don't show in
 * the admin tool service list.
 *
 * Revision 1.6  2002/03/17 16:15:26  chris
 * Do not overwrite GWCONFIG if it already exists.
 *
 * Revision 1.5  2002/03/17 11:54:21  chris
 * New config compile options
 *
 * Revision 1.4  2002/02/17 23:46:25  chris
 * Remove flattree.h
 *
 * Revision 1.3  2002/01/28 20:45:16  chris
 * Creates binary config file.
 *
 * Revision 1.2  2002/01/20 21:18:46  chris
 * Builds regtree from xml config file.
 *
 * Revision 1.1  2002/01/19 11:48:28  chris
 * New config example and ascii->binary compiler for it.
 *
 * --------------------------------------------------------------------------
 */
#include <stdio.h>
#include <expat.h>

#include <phlib/xl_mmap.h>
#include <phlib/xl_linklst.h>

#include "gwgbprm.h"
#include "gwlog.h"

#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "compileconf.h"
#include "conftree.h"

#include <expat.h>

#ifdef DEBUG
# define IF_DEBUG if(1)
#else
# define IF_DEBUG if(0)
#endif

#define PADDING   "                "
#define SIGNATURE "GWCFG-1.4   xxxx"

static int32 endian_check = 0xaabbccdd;

/* -- Global Context Status Values -- */
static int indent = 0;

static XMLTag XML_Tag_Stack[ MAX_TAG_DEPTH ];
static long   XML_SP    = 0;
static long   xml_level = 0;
static int    last_branch_end[64];

static uchar_t tag_usage[ MAX_TOKEN ];

typedef struct {
  long remote_dom_count;
  long min;
  long max;
  long priority;
} Globals;

static Globals glob;

static CString sepCS           = CSInit;
static CString branchCS        = CSInit;

/* Declare how many repeat request row 'holder' lists we 'cache' between
 * requests...
 */
/* static int REPEAT_REQ_ROW_MAX_CACHE = 10;*/

LIST *FuncConfLst = NULL;

#define ERROR_IF_RECURSIVE(a,b) if( tag_usage[(a)] > 1 ) { \
        printf( "<%s> may not be used recursively\n", (b) ); \
        return -1; }

/* Prototypes */
static int  tag_handler    ( const char *tag, int flags, 
                             long data_SP, char **attr );
static int  process_tag    ( const char *tag, TagToken tagid, int flags );
static int  close_tag      ( XMLTag *xmltag );

static void TagStartHandler( void *udata, const char *tag, const char **attr );
static void TagEndHandler  ( void *udata, const char *tag );
static void TagDataHandler ( void *udata, const char *data, int len );

int strcmp_ci   ( const char *str1, const char     *str2 );
int stack_search( const char *tag,  const TokenMap *     );
int tokenise    ( const char *tag,  const TokenMap *map  );

int element_is_array( TagToken token );
int element_is_leaf( TagToken token );
int fn_import_xml_config( const char *fname );


/* ------------------------------------------------------------------------- */

int
main( int argc, char **argv )
{
  struct stat sb = {0};
  CString memCS = CSInit;
  int        fd = 0;
  int allow_dump = 0;
  int fname_idx  = 1;
  char *gwconf   = "";
  
  if( (gwconf = getenv( "GWCONFIG" )) == NULL ) {
    printf( "GWCONFIG environment variable not set\n" );
    exit(1);
  }

  if( argc == 3 && strcmp( argv[fname_idx], "--dump" ) == 0 ) { 
    allow_dump = 1; 
	fname_idx++;
	argc--;
  };

  if( argc != 2 ) {
    printf( "usage: gwmkcfg <config file>\n" );
    exit(2);
  }


  start_tree();

  printf( "Goldwater Configuration Compiler - v1.4 - (c)2003 netFluid Technology.\n" );

  if( fn_import_xml_config( argv[fname_idx] ) < 0 ) {
    printf( "Cannot process input configuration file '%s'\n", argv[fname_idx] );
    exit(3);
  }

  if( allow_dump ) dump_tree();

  flatten_tree( &memCS );

  if( stat( gwconf, &sb ) == 0 ) {
    if( S_ISREG(sb.st_mode) ) {
	  char c = 0;
	  /* File already exists */
	  printf( "\n%s already exists!\n(Warning Goldwater may be running and accessing this file!)\n\nDo you wish to overwrite file? [n] ", gwconf );
      c = getchar();
	  if( c != 'Y' && c != 'y' ) {
	    printf( "\nExiting\n" );
		exit(1);
	  }
	} else {
	  /* Probably a bad file */
	  printf( "'%s' is not a regular file\n", gwconf );
	  exit(3);
	}
  } 

  if( (fd = open( gwconf, O_WRONLY | O_CREAT | O_TRUNC, 0666 )) < 0 ) {
    printf( "Cannot open config file '%s' for writing\n", gwconf );
  }

  write( fd, SIGNATURE PADDING, 12 );
  write( fd, &endian_check, 4      ); /* Write out endian check */

  write( fd, CSStart(memCS), CSLen(memCS) );
  close(fd);

  printf( "\n%s created - %d bytes\n", gwconf, CSLen(memCS) );
  

  free_cstring( &memCS );

  end_tree();

  exit(0);
}

/* ------------------------------------------------------------------------- */

int
fn_import_xml_config( const char *fname )
{
  void   *mem     = NULL;
  struct  stat sb = {0};
  int     fd      = -1;
  int     ret     = 0;
  CString xmlCS   = CSInit;

  XML_Parser xc;

  /* Sort out constants
   */
  WrapCStringLen( sepCS, "/", 1 );

  free_cstring( &branchCS );


  if( (fd = open( (char*)fname, O_RDONLY ) ) < 0 ) ret = (-3);
  if( ret == 0 && fstat( fd, &sb ) < 0 ) ret = (-2);

  if( ret == 0 ) {
    if( (mem = mmap( NULL, (int)sb.st_size,
		     PROT_READ, MAP_PRIVATE, fd, 0 )) >= (void *)0 ) {
      WrapCStringPtrLen( &xmlCS, (char*)mem, (int)sb.st_size );
    } else ret = (-1);
  }

  if( ret > (-3) ) close( fd );
  if( ret < 0 ) return ret;

  /* Begin XML processing
   */
  indent    = 0;
  XML_SP    = 0;
  xml_level = 0;

  memset( tag_usage, 0, sizeof( tag_usage ) ); /* Clear Tag Usage */

  memset( &glob, 0, sizeof(Globals) ); /* Clear globals */
  
  xc = XML_ParserCreate(NULL);  /* Initialise XML Engine */

  /* Register callbacks for tags etc
   */
  XML_SetElementHandler      ( xc, TagStartHandler, TagEndHandler );
  XML_SetCharacterDataHandler( xc, TagDataHandler );

  /* Pass XML into parser (done set to 1 as there is no more after this) 
   */
  if( ! XML_Parse( xc, CSStart(xmlCS), CSLen(xmlCS), 1 ) ) {
    printf( "Error processing XML line %d: %s\n",
	    XML_GetCurrentLineNumber(xc), 
	    XML_ErrorString(XML_GetErrorCode(xc)));
    return -1;
  }

  /* Free memory used by parser */
  XML_ParserFree( xc );
  
  /* Release Config XML File */
  munmap( mem, (int)sb.st_size );

  return 0;
}

/* ------------------------------------------------------------------------- */
/* Callback Function - Registered with xml_parser
 */
static void
TagStartHandler( void *udata, const char *tag, const char **attr )
{
  int     ix     = 0;
  CString nodeCS = CSInit;

  indent += 2;

  last_branch_end[ xml_level ] = CSLen( branchCS );
  xml_level++;
  
  /* Add new node to end of branch */
  WrapCString( nodeCS, tag );
  append_cstring( &branchCS, &sepCS  );
  append_cstring( &branchCS, &nodeCS );

  IF_DEBUG {
    printf( "%*.*sStart Tag: '%s'\n", indent, indent, "", tag );
    for( ix = 0; attr[ix]; ix += 2 ) 
      printf( "%*.*s + Param %02d: %s = '%s'\n",
	      indent,indent,"",ix/2,attr[ix],attr[ix+1]);
  }

  if( tag_handler( tag, F_TAG_OPEN, XML_SP, (char**)attr ) < 0 ) {
    printf( "tag_handler failed. (tag %s)\n", tag );
    exit(1);
  }
}

/* ------------------------------------------------------------------------- */
/* Callback Function - Registered with xml_parser
 */
static void
TagEndHandler( void *udata, const char *tag )
{
  IF_DEBUG printf( "%*.*sEnd Tag: '%s'\n", indent, indent, "", tag );
  indent -= 2;

  xml_level--;

  if( tag_handler( tag, F_TAG_CLOSE, XML_SP, NULL ) < 0 ) {
    printf( "tag_handler (closing) failed. (tag %s)\n", tag );
    exit(1);
  }

  /* Restore branch */
  CSLen( branchCS ) = last_branch_end[ xml_level ];
}

/* ------------------------------------------------------------------------- */
/* Callback Function - Registered with xml_parser
 */
static void
TagDataHandler( void *udata, const char *data, int len )
{
  CString dataCS = CSInit;
  long    dataSP = 0;

  WrapCStringLen( dataCS, (char*)data, len );

  /* Strip trailing \n, \r and space characters etc */
  while( CSLen(dataCS) > 0 &&
	 CSChar( dataCS, CSLen(dataCS) - 1 ) <= ' ' ) CSLen(dataCS) -= 1;
  
  /* Strip leading spaces */
  skip_lead_space( &dataCS );

  if( CSLen( dataCS ) <= 0 ) return;

  IF_DEBUG printf( "%*.*s * Data[%d]: '%*.*s'\n", indent,indent,"", 
		   CSLen(dataCS),
		   CSLen(dataCS),
		   CSLen(dataCS),
		   CSStart(dataCS) );

  dataSP = XML_Tag_Stack[ XML_SP ].data_SP;

  IF_DEBUG printf( "Associating data with tagid %d\n", 
		   XML_Tag_Stack[ dataSP ].token );

  append_cstring( &XML_Tag_Stack[ dataSP ].dataCS, &dataCS );
}

/* ------------------------------------------------------------------------- */
/*
 * This handler function is recurrsive to handle embedded tag modifiers
 * Eg default in <tag default=" ">
 */
int
tag_handler( const char *tag, int flags, long data_SP, char **attr )
{
  int tagid = TAG_UNKNOWN;
  int ix    = 0;
  int iy    = 0;
  int ret   = 0;

  if( (tagid = tokenise( tag, Mapping )) == TAG_UNKNOWN ) {
    printf( "unknown XML tag in config '%s'\n", tag );
    return -1;
  }

  /* write(1, ".", 1); */

  if( flags & F_TAG_OPEN ) {
    /* We're opening a new tag
     */
    XML_SP++;
    
    XML_Tag_Stack[ XML_SP ].token    = tagid;
    XML_Tag_Stack[ XML_SP ].balanced = 1;
    XML_Tag_Stack[ XML_SP ].data_SP  = data_SP + 1;
    WrapCStringLen( XML_Tag_Stack[ XML_SP ].dataCS, "", 0 );
    WrapCStringLen( XML_Tag_Stack[ XML_SP ].elementCS, "", 0 );
    
    append_chars( &XML_Tag_Stack[ XML_SP ].elementCS, tag );

    if( process_tag( tag, tagid, flags ) < 0 ) return -1;

    tag_usage[ tagid ] ++;
    
    /* Process parameters and assign them to the main tag as associated tags
     * that need auto-closing on closing this tag.
     *
     * data_SP forces all freetext data to be associated with the top level
     * tag. eg <foo bar="value">freetext</foo> 
     * creates a stack containing 'bar' at the top, associated with it 'value'
     * and 'foo' below it.  When we hit 'freetext' we would naturally assoc
     * it with the tag at the top of the stack, 'bar' in this case.  This is
     * incorrect, the text should be associated with 'foo'.  We can infer this
     * easily, hence the use of data_SP below to force the correct association.
     */
    if( attr ) {
      for( ix = 0; attr[ix]; ix += 2 ) {
        if( tag_handler( attr[ix], F_TAG_OPEN | F_TAG_EMBEDDED, 
		         data_SP, NULL ) < 0 ) return -1;
        /* These are auto-tags, so we don't need to do a balance check when
         * closing the top-level tag. 
         */
        XML_Tag_Stack[ XML_SP ].balanced = 0;
        
        if( attr[ix+1] > 0 ) 
	  append_chars( &XML_Tag_Stack[ XML_SP ].dataCS, attr[ix+1] );
      }
    }
  }

  if( flags & F_TAG_CLOSE ) {
    /* We need to close the requested tag, and any others within its stack
     * frame.
     */
    
    /* Scan back through stack, check for balance...
     */
    for( ix = XML_SP; ix > 0; ix-- ) {
      if( XML_Tag_Stack[ ix ].token == tagid ) break;
      if( XML_Tag_Stack[ ix ].balanced ) {
	printf( "XML is unbalanced.  Tags are not closed properly.\n" );
	return -1;
      }
    }

    if( ix == 0 ) {
      printf( "XML is unbalanced.  Tags are not closed properly.\n" );
      return -1;
    }
    
    /* Matching open tag found... close tags above it on the stack
     */
    for( iy = XML_SP; iy >= ix; iy-- ) {
      ret = close_tag( &XML_Tag_Stack[ iy ] );
      
      tag_usage[ XML_Tag_Stack[ iy ].token ] --;
      
      free_cstring( &XML_Tag_Stack[ iy ].dataCS );
      free_cstring( &XML_Tag_Stack[ iy ].elementCS );
      XML_SP--;

      if( ret < 0 ) return -1;
    }
  }
  return ret;
}

/* ------------------------------------------------------------------------- */
/*
 * Search back through the stack for tag matching 'tag'
 * Return -1 if not found, else stack location.
 */
int
stack_search( const char *tag, const TokenMap *map )
{
  int tagid = TAG_UNKNOWN;
  int ix    = 0;

  if( (tagid = tokenise( tag, map )) == TAG_UNKNOWN ) {
    printf( "Unknown tag '%s'\n", tag );
    return -1;
  }

  /* Scan back through stack, check for balance...
   */
  for( ix = XML_SP; ix > 0; ix-- ) {
    if( XML_Tag_Stack[ ix ].token == tagid ) break;
  }

  return (ix > 0) ? ix : -1;
}

/* ------------------------------------------------------------------------- */
/*
 * Validate XML tag - does it appear in the correct context?
 */
static int
validate_tag( const char *tag, TagToken tagid, int flags )
{
  /* Check depth etc */
  switch( tagid ) {

  case T_GOLDWATER:
    ERROR_IF_RECURSIVE(tagid,tag);
    break;

  case T_DOMAIN:
  case T_CONTROL:
    ERROR_IF_RECURSIVE(tagid,tag);
    if( tag_usage[ T_GOLDWATER ] == 0 || xml_level != 2 ) {
      printf( "<%s> may only be used within a <goldwater>..</goldwater> scope\n", tag );
      return -1;
    }
    break;

  case T_NS:
  case T_ADDR:
  case T_PORT:
  case T_EXEDIR:
  case T_SANDBOX:
  case T_LOGFILE:
  case T_LOGLEVEL:
  case T_SPOOLDIR:
  case T_MODE:
  case T_IPCKEY:
  case T_REQTOUT:
  case T_MAXSERVERS:
  case T_MAXSERVICES:
  case T_MAXCLIENTS:
  case T_USER:
  case T_GROUP:
  case T_UMASK:
    ERROR_IF_RECURSIVE(tagid,tag);
    if( tag_usage[ T_DOMAIN ] == 0 || xml_level != 2 ) {
      printf( "<%s> may only be used within a <domain>..</domain> scope\n", tag );
      return -1;
    }
    break;

  case T_IMPORT:
    if( tag_usage[ T_DOMAIN ] == 0 || xml_level != 3 ) {
      printf( "<%s> may only be used within a <domain>..</domain> scope\n", tag );
      return -1;
    }
    break;

  case T_MAXDOMAINS:
  case T_LOCAL:
  case T_REMOTE:
    ERROR_IF_RECURSIVE(tagid,tag);
    if( tag_usage[ T_CONTROL ] == 0 || xml_level != 3 ) {
      printf( "<%s> may only be used within a <control>..</control> scope\n", tag );
      return -1;
    }
    break;

  case T_DEFAULT:
  case T_SERVERS:
  case T_ENVIRONMENT:
    ERROR_IF_RECURSIVE(tagid,tag);
    if( tag_usage[ T_DOMAIN ] == 0 || xml_level != 3 ) {
      printf( "<%s> may only be used within a <domain>..</domain> scope\n", tag );
      return -1;
    }
    break;

  case T_MAX:
  case T_MIN:
  case T_MAXREQUESTS:
  case T_PRIORITY:
    ERROR_IF_RECURSIVE(tagid,tag);
    if( (xml_level == 3 && tag_usage[ T_DEFAULT ] == 0 ) ||
	(xml_level == 4 && tag_usage[ T_CLASS   ] == 0 ) ) {
      printf( "<%s> may only be used within a <class>..</class> scope\n", tag );
      return -1;
    }
    break;
    
  case T_EXE:
    ERROR_IF_RECURSIVE(tagid,tag);
    if( tag_usage[ T_CLASS ] == 0 || xml_level != 4 ) {
      printf( "<%s> may only be used within a <class>..</class> scope\n", tag );
      return -1;
    }
    break;

  case T_SERVICE:
    ERROR_IF_RECURSIVE(tagid,tag);
    
    if( tag_usage[ T_CLASS ] == 0 || xml_level != 5 ) {
      printf( "<%s> may only be used within a <class>..</class> scope\n", tag );
      return -1;
    }
    break;

  case T_FUNCTION:
  case T_EXPORT:
    ERROR_IF_RECURSIVE(tagid,tag);
    
    if( tag_usage[ T_SERVICE ] == 0 || xml_level != 5 ) {
      printf( "<%s> may only be used within a <service>..</service> scope\n", tag );
      return -1;
    }
    break;

  default:
    break;
  }

  /* printf( "XML Level: %d\n", xml_level ); */
  return 0;
}


/* ------------------------------------------------------------------------- */
/* Called on discovering a close tag
 */
static int
close_tag( XMLTag *xmltag )
{
  int              ret       = 0;
  CString          Ascii0CS  = CSInit;
  int              blen       = 0;

  WrapCStringLen( Ascii0CS, "\0", 1 );

  /* Handles cleanup of state and processes any data in xmltag that has not
   * been delt with.
   */
  IF_DEBUG {
    printf( "Close tag id %d (%*.*s)\n", xmltag->token,
	    CSLen  (xmltag->elementCS),
	    CSLen  (xmltag->elementCS),
	    CSStart(xmltag->elementCS) );
    
    printf( "        data %*.*s\n", 
	    CSLen  (xmltag->dataCS),
	    CSLen  (xmltag->dataCS),
	    CSStart(xmltag->dataCS) );
  }
  
#if 0
  /* Validate data (if a number, make sure it is a number */
  /* CStoInt32( &xmltag->dataCS, (int32*)&tmpval ); */
  if( validate_data( xmltag ) == Failure ) {
    printf( "Invalid numeric value: %*.*s\n", CSLen(xmltag->dataCS),
	    CSLen(xmltag->dataCS), CSStart(xmltag->dataCS) );
    return -1;
  }
#endif
  
  skip_lead_space( &xmltag->dataCS );
  skip_tail_space( &xmltag->dataCS );

  if( element_is_leaf( xmltag->token ) || CSLen( xmltag->dataCS ) ) {
    /* If the node really is a leaf node, then we assign data straight to
     * it.  If it's a branch that also has data ( <el val="123">456</el> )
     * then we create a 'false' node el=123.
     */
    blen = CSLen( branchCS );
    append_cstring( &branchCS, &sepCS );
    append_cstring( &branchCS, &xmltag->elementCS );
    append_cstring( &branchCS, &Ascii0CS );
    CSLen( branchCS ) = blen; /*The added node is temporary! */
    
    IF_DEBUG printf( "Add Leaf: '%s'\n", CSStart(branchCS) );
    reg_add_leaf( CSStart( branchCS ),
		  CSStart(xmltag->dataCS), CSLen(xmltag->dataCS), ListEnd );
  }

  return ret;
}

/* ------------------------------------------------------------------------- */

int
element_is_array( TagToken token )
{
  int ret = 0;

  switch( token ) {
  case T_ENVIRONMENT:
  case T_SERVICE:
  case T_CLASS:
  case T_REMOTE:
  case T_DOMAIN:
  case T_IMPORT:
    ret = 1;
    break;
    
  default:
    ret = 0;
  }

  return ret;
}

/* ------------------------------------------------------------------------- */

int
element_is_leaf( TagToken token )
{
  switch( token ) {
    /* Branch Tokens */
  case T_GOLDWATER:
  case T_CONTROL:
  case T_DOMAIN:
  case T_DEFAULT:
  case T_SERVERS:
  case T_SERVICE:
  case T_CLASS:
    return 0;
    break;

    /* Leaf Tokens */
  case T_LOCAL:
  case T_REMOTE:
  case T_NAME:
  case T_FUNCTION:
  case T_EXPORT:
  case T_NS:
  case T_ADDR:
  case T_PORT:
  case T_EXEDIR:
  case T_SANDBOX:
  case T_USER:
  case T_GROUP:
  case T_UMASK:
  case T_LOGFILE:
  case T_LOGLEVEL:
  case T_SPOOLDIR:
  case T_MODE:
  case T_IPCKEY:
  case T_REQTOUT:
  case T_MAXSERVERS:
  case T_MAXCLIENTS:
  case T_MAXSERVICES:
  case T_MIN:
  case T_MAX:
  case T_MAXREQUESTS:
  case T_PRIORITY:
  case T_ID:
  case T_EXE:
  case T_MAXDOMAINS:
    return 1;
    break;

  default:
    return 0;
  }

  return 0;
}

/* ------------------------------------------------------------------------- */
/*
 * Called on discovering an opening tag
 */
static int
process_tag( const char *tag, TagToken tagid, int flags )
{
  int  l;
  char array_node[20];
  CString tmpCS = CSInit;
  
  IF_DEBUG printf( "Processing tag id %d (%s)\n", tagid, tag ); 
  
  if( validate_tag( tag, tagid, flags ) < 0 ) return -1;

  /* If element is an array, then we need to create a intermediate node
   * giving the array index.
   */
  if( element_is_array( tagid ) ) {
    
    /* printf( "COUNT CHILDREN OF '%*.*s'\n",
       CSLen(branchCS), CSLen(branchCS), CSStart(branchCS) ); */
    
    l = sprintf( array_node, "/%06d", count_children( &branchCS ) );
    WrapCStringLen( tmpCS, array_node, l );
    append_cstring( &branchCS, &tmpCS );
  }
  
  IF_DEBUG printf( "Branch: '%*.*s'\n", CSLen(branchCS), CSLen(branchCS), CSStart(branchCS) );
  
  return 0;
}

/* ------------------------------------------------------------------------- */

int
strcmp_ci( const char *str1, const char *str2)
{
  int res, i, len1;
  char c1, c2;

  res = 0;
  len1 = strlen( str1 );
  for(i = 0; i < len1 && res == 0; i++) {
    c1 = str1[i];
    if( (c2 = str2[i]) == 0 ) return 1; /* str2 shorter than str1 */

    if(c1 >= 'a' && c1 <= 'z') c1 -= 32;
    if(c2 >= 'a' && c2 <= 'z') c2 -= 32;

    if(c1 < c2) res = MINUSONE;
    else if(c1 > c2) res = 1;
  }
  if( res == 0 && str2[i] != 0 ) res = -1; /* str2 is longer than str1 */

  return res;
}

/* ------------------------------------------------------------------------- */

int
tokenise( const char *tag, const TokenMap *map )
{
  int   tkn = TAG_UNKNOWN;
  char *txt = NULL; 
  int   i   = 0; 
 
  /* Match to token table */
  while( (txt = map[i].t_text) != NULL ) { 
    if( strcmp_ci( txt, tag ) == 0 ) {
      tkn = map[i].token;
      break;
    }
    i++;
  }
  return tkn;
}

/* ------------------------------------------------------------------------- */
/* end */


